from typing import List, Tuple
import itertools
from copy import copy

from asf_search.ASFSearchOptions import ASFSearchOptions
from asf_search.constants import CMR_PAGE_SIZE
from asf_search.CMR.datasets import (
    NISAR_PRODUCT_TYPES,
    collections_by_processing_level,
    collections_per_platform,
    get_concept_id_alias,
    get_dataset_concept_ids,
)
from numpy import intersect1d, union1d

def build_subqueries(opts: ASFSearchOptions) -> List[ASFSearchOptions]:
    """
    Build a list of sub-queries using the cartesian product
    of all the list parameters described by opts

    :param opts: The search options to split into sub-queries
    :return list: A list of ASFSearchOptions objects
    """
    params = dict(opts)

    # Break out two big list offenders into manageable chunks
    for chunked_key in ['granule_list', 'product_list']:
        if params.get(chunked_key) is not None:
            params[chunked_key] = chunk_list(params[chunked_key], CMR_PAGE_SIZE)

    list_param_names = [
        'platform',
        'season',
        'collections',
        'dataset',
        'instrument',
        'cmr_keywords',
        'shortName',
        'circle',
        'linestring',
        'point',
        'bbox',
    ]  # these parameters will dodge the subquery system
    skip_param_names = [
        'maxResults',
    ]  # these params exist in opts, but shouldn't be passed on to subqueries at ALL

    includes_nisar_products = False
    if params.get('processingLevel') is not None:
        for product in params.get('processingLevel', []):
            if product in NISAR_PRODUCT_TYPES:
                includes_nisar_products = True
                break
    collections, aliased_keywords = get_keyword_concept_ids(params, opts.collectionAlias, includes_nisar_products)
    params['collections'] = list(union1d(collections, params.get('collections', [])))

    for keyword in [*skip_param_names, *aliased_keywords]:
        params.pop(keyword, None)

    subquery_params, list_params = {}, {}
    for key, value in params.items():
        if key in list_param_names:
            list_params[key] = value
        else:
            subquery_params[key] = value

    sub_queries = cartesian_product(subquery_params)
    return [_build_subquery(query, opts, list_params) for query in sub_queries]


def _build_subquery(
    query: List[Tuple[dict]], opts: ASFSearchOptions, list_params: dict
) -> ASFSearchOptions:
    """
    Composes query dict and list params into new ASFSearchOptions object

    param: query: the cartesian search query options
    param: opts: the search options to pull config options from (provider, host, session)
    param: list_params: the subquery parameters
    """
    q = dict()
    for p in query:
        q.update(p)

    q['provider'] = opts.provider
    q['host'] = opts.host
    q['session'] = copy(opts.session)

    return ASFSearchOptions(**q, **list_params)


def get_keyword_concept_ids(params: dict, use_collection_alias: bool = True, includes_nisar_products: bool = False) -> dict:
    """
    Gets concept-ids for dataset, platform, processingLevel keywords
    processingLevel is scoped by dataset or platform concept-ids when available

    : param params:
        search parameter dictionary pre-CMR translation
    : param use_collection_alias:
        whether or not to alias platform and processingLevel with concept-ids
    : includes_nisar_products:
        Flag to skip the processing level aliasing (urgent response products are grouped by product level and not product type)
    : returns two lists:
        - list of concept-ids for dataset, platform, and processingLevel
        - list of aliased keywords to remove from final parameters
    """
    collections = []
    aliased_keywords = []

    if use_collection_alias:
        alias_processing_levels = 'processingLevel' in params.keys() and not includes_nisar_products
        if alias_processing_levels:
            collections = get_concept_id_alias(
                params.get('processingLevel'), collections_by_processing_level
            )
            if len(collections):
                aliased_keywords.append('processingLevel')

        alias_platforms = 'platform' in params.keys()
        if alias_platforms:
            platform_concept_ids = get_concept_id_alias(
                [platform.upper() for platform in params.get('platform')],
                collections_per_platform,
            )
            if len(platform_concept_ids):
                aliased_keywords.append('platform')
                collections = _get_intersection(platform_concept_ids, collections)
        if alias_processing_levels and alias_platforms and not len(collections):
            # processingLevel isn't part of the given platform. Drop aliased keywords so we don't search on zero collections
            aliased_keywords = []
            
    if 'dataset' in params.keys():
        aliased_keywords.append('dataset')
        dataset_concept_ids = get_dataset_concept_ids(params.get('dataset'))
        collections = _get_intersection(dataset_concept_ids, collections)

    return collections, aliased_keywords


def _get_intersection(keyword_concept_ids: List[str], intersecting_ids: List[str]) -> List[str]:
    """
    Returns the intersection between two lists. If the second list is empty the first list
    is return unchaged
    """
    if len(intersecting_ids):
        return list(intersect1d(intersecting_ids, keyword_concept_ids))

    return keyword_concept_ids


def chunk_list(source: List, n: int) -> List:
    """
    Breaks a longer list into a list of lists, each of length n

    :param source: The list to be broken into chunks
    :param n: The maximum length of each chunk

    :return List[List, ...]:
    """
    return [source[i * n : (i + 1) * n] for i in range((len(source) + n - 1) // n)]


def cartesian_product(params):
    formatted_params = format_query_params(params)
    p = list(itertools.product(*formatted_params))
    return p


def format_query_params(params) -> List[List[dict]]:
    listed_params = []

    for param_name, param_val in params.items():
        plist = translate_param(param_name, param_val)
        listed_params.append(plist)

    return listed_params


def translate_param(param_name, param_val) -> List[dict]:
    param_list = []

    if not isinstance(param_val, list):
        param_val = [param_val]

    for unformatted_val in param_val:
        formatted_val = unformatted_val

        if isinstance(unformatted_val, list):
            formatted_val = ','.join([f'{t}' for t in unformatted_val])

        param_list.append({param_name: formatted_val})

    return param_list



import os
from typing import Any, Dict, Tuple, Type, List, final
import warnings
from shapely.geometry import shape, Point, Polygon, mapping
import json
import re

from urllib import parse

from asf_search import ASFSession, ASFSearchResults
from asf_search.ASFSearchOptions import ASFSearchOptions
from asf_search.download import download_url
from asf_search.download.file_download_type import FileDownloadType
from asf_search.CMR.translate import try_parse_date
from asf_search.CMR.translate import try_parse_float, try_parse_int, try_round_float


class ASFProduct:
    """
    The ASFProduct class is the base class for search results from asf-search.
    Key props:
        - properties:
            - stores commonly acessed properties of the CMR UMM for convenience
        - umm:
            - The data portion of the CMR response
        - meta:
            - The metadata portion of the CMR response
        - geometry:
            - The geometry `{coordinates: [[lon, lat] ...], 'type': Polygon}`
        - baseline:
            - used for spatio-temporal baseline stacking, stores state vectors/ascending
                node time/insar baseline values when available (Not set in base ASFProduct class)
            - See `S1Product` or `ALOSProduct` `get_baseline_calc_properties()`
                methods for implementation examples

    Key methods:
        - `download()`
        - `stack()`
        - `remotezip()`


    """

    @classmethod
    def get_classname(cls):
        return cls.__name__

    _base_properties = {
        # min viable product
        'centerLat': {
            'path': ['AdditionalAttributes', ('Name', 'CENTER_LAT'), 'Values', 0],
            'cast': try_parse_float,
        },
        'centerLon': {
            'path': ['AdditionalAttributes', ('Name', 'CENTER_LON'), 'Values', 0],
            'cast': try_parse_float,
        },
        'stopTime': {
            'path': ['TemporalExtent', 'RangeDateTime', 'EndingDateTime'],
            'cast': try_parse_date,
        },  # primary search results sort key
        'fileID': {'path': ['GranuleUR']},  # secondary search results sort key
        'flightDirection': {
            'path': [
                'AdditionalAttributes',
                ('Name', 'ASCENDING_DESCENDING'),
                'Values',
                0,
            ]
        },
        'pathNumber': {
            'path': ['AdditionalAttributes', ('Name', 'PATH_NUMBER'), 'Values', 0],
            'cast': try_parse_int,
        },
        'processingLevel': {
            'path': ['AdditionalAttributes', ('Name', 'PROCESSING_TYPE'), 'Values', 0]
        },
        # commonly used
        'url': {'path': ['RelatedUrls', ('Type', 'GET DATA'), 'URL']},
        'startTime': {
            'path': ['TemporalExtent', 'RangeDateTime', 'BeginningDateTime'],
            'cast': try_parse_date,
        },
        'sceneName': {
            'path': [
                'DataGranule',
                'Identifiers',
                ('IdentifierType', 'ProducerGranuleId'),
                'Identifier',
            ]
        },
        'browse': {'path': ['RelatedUrls', ('Type', [('GET RELATED VISUALIZATION', 'URL')])]},
        'platform': {'path': ['AdditionalAttributes', ('Name', 'ASF_PLATFORM'), 'Values', 0]},
        'bytes': {
            'path': ['AdditionalAttributes', ('Name', 'BYTES'), 'Values', 0],
            'cast': try_round_float,
        },
        'md5sum': {'path': ['AdditionalAttributes', ('Name', 'MD5SUM'), 'Values', 0]},
        'frameNumber': {
            'path': ['AdditionalAttributes', ('Name', 'CENTER_ESA_FRAME'), 'Values', 0],
            'cast': try_parse_int,
        },  # overloaded by S1, ALOS, and ERS
        'granuleType': {'path': ['AdditionalAttributes', ('Name', 'GRANULE_TYPE'), 'Values', 0]},
        'orbit': {
            'path': ['OrbitCalculatedSpatialDomains', 0, 'OrbitNumber'],
            'cast': try_parse_int,
        },
        'polarization': {'path': ['AdditionalAttributes', ('Name', 'POLARIZATION'), 'Values', 0]},
        'processingDate': {
            'path': ['DataGranule', 'ProductionDateTime'],
            'cast': try_parse_date,
        },
        'sensor': {
            'path': ['Platforms', 0, 'Instruments', 0, 'ShortName'],
        },
    }
    """
    _base_properties dictionary, mapping readable property names to paths and optional type casting

    entries are organized as such:
        - `PROPERTY_NAME`: The name the property should be called in `ASFProduct.properties`
            - `path`: the expected path in the CMR UMM json granule response as a list
            - `cast`: (optional): the optional type casting method

    Defining `_base_properties` in subclasses allows for
    defining custom properties or overiding existing ones.
    See `S1Product.get_property_paths()` on how subclasses are expected to
    combine `ASFProduct._base_properties` with their own separately defined `_base_properties`
    """

    _url_types = ['GET DATA', 'EXTENDED METADATA', 'GET DATA VIA DIRECT ACCESS', 'GET RELATED VISUALIZATION', 'VIEW RELATED INFORMATION', 'USE SERVICE API']
    
    def __init__(self, args: Dict = {}, session: ASFSession = ASFSession()):
        self.meta = args.get('meta')
        self.umm = args.get('umm')

        translated = self.translate_product(args)

        self.properties = translated['properties']
        self.geometry = translated['geometry']
        self.baseline = None
        self.session = session

    def __str__(self):
        return json.dumps(self.geojson(), indent=2, sort_keys=True)

    def geojson(self) -> Dict:
        """
        Returns ASFProduct object as a geojson formatted dictionary
        with `type`, `geometry`, and `properties` keys
        """
        return {
            'type': 'Feature',
            'geometry': self.geometry,
            'properties': self.properties,
        }

    def download(
        self,
        path: str,
        filename: str = None,
        session: ASFSession = None,
        fileType=FileDownloadType.DEFAULT_FILE,
    ) -> None:
        """
        Downloads this product to the specified path and optional filename.

        :param path: The directory into which this product should be downloaded.
        :param filename: Optional filename to use instead of the original filename of this product.
        :param session: The session to use, defaults to the one used to find the results.

        :return: None
        """

        default_filename = self.properties['fileName']

        if filename is not None:
            multiple_files = (
                fileType == FileDownloadType.ADDITIONAL_FILES
                and len(self.properties['additionalUrls']) > 1
            ) or fileType == FileDownloadType.ALL_FILES
            if multiple_files:
                warnings.warn(
                    'Attempting to download multiple files for product, '
                    f'ignoring user provided filename argument "{filename}", using default.'
                )
            else:
                default_filename = filename

        if session is None:
            session = self.session

        urls = self.get_urls(fileType=fileType)

        for url in urls:
            base_filename = '.'.join(default_filename.split('.')[:-1])
            extension = url.split('.')[-1]
            download_url(
                url=url,
                path=path,
                filename=f'{base_filename}.{extension}',
                session=session,
            )

    def get_urls(self, fileType=FileDownloadType.DEFAULT_FILE) -> list:
        urls = []

        if fileType == FileDownloadType.DEFAULT_FILE:
            urls.append(self.properties['url'])
        elif fileType == FileDownloadType.ADDITIONAL_FILES:
            urls.extend(self.properties.get('additionalUrls', []))
        elif fileType == FileDownloadType.ALL_FILES:
            urls.append(self.properties['url'])
            urls.extend(self.properties.get('additionalUrls', []))
        else:
            raise ValueError(
                "Invalid FileDownloadType provided, the valid types are 'DEFAULT_FILE', 'ADDITIONAL_FILES', and 'ALL_FILES'"
            )
        return urls

    def _get_additional_filenames_and_urls(
        self,
        default_filename: str = None,  # for subclasses without fileName in url (see S1BurstProduct implementation) # noqa F401
    ) -> List[Tuple[str, str]]:
        return [
            (self._parse_filename_from_url(url), url)
            for url in self.properties.get('additionalUrls', [])
        ]

    def _parse_filename_from_url(self, url: str) -> str:
        file_path = os.path.split(parse.urlparse(url).path)
        filename = file_path[1]
        return filename

    def stack(
        self, opts: ASFSearchOptions = None, useSubclass: Type['ASFProduct'] = None
    ) -> ASFSearchResults:
        """
        Builds a baseline stack from this product.
        Parameters
        ----------
        opts:
        An ASFSearchOptions object describing the search parameters to be used.
        Search parameters specified outside this object will override in event of a conflict.
        ASFProductSubclass: An ASFProduct subclass constructor to cast results to

        Returns
        ----------
        asf_search.ASFSearchResults
            containing the stack, with the addition of baseline values
            (temporal, perpendicular) attached to each ASFProduct.
        """
        from .search.baseline_search import stack_from_product

        if opts is None:
            opts = ASFSearchOptions(session=self.session)

        return stack_from_product(self, opts=opts, ASFProductSubclass=useSubclass)

    def get_stack_opts(self, opts: ASFSearchOptions = None) -> ASFSearchOptions:
        """
        Build search options that can be used to find an insar stack for this product

        :return: ASFSearchOptions describing appropriate options
        for building a stack from this product
        """
        return None

    def _get_access_urls(
        self, 
        url_types: List[str] = ['GET DATA', 'EXTENDED METADATA']
    ) -> List[str]:
        accessUrls = []

        for url_type in url_types:
            if urls := self.umm_get(self.umm, 'RelatedUrls', ('Type', [(url_type, 'URL')]), 0):
                accessUrls.extend(urls)

        return sorted(list(set(accessUrls)))

    def _get_urls(self) -> List[str]:
        """Finds and returns all umm urls"""
        urls = self._get_access_urls(self._url_types)
        return [
            url for url in urls if not url.startswith('s3://')
        ]

    def _get_s3_uris(self) -> List[str]:
        """Finds and returns all umm S3 direct access uris"""
        s3_urls = self._get_access_urls(self._url_types)
        return [url for url in s3_urls if url.startswith('s3://')]

    def _get_additional_urls(self) -> List[str]:
        """Finds and returns all non-md5/image urls and filters out the existing `url` property"""
        access_urls = self._get_urls()
        return [
            url for url in access_urls
            if not url.endswith('.md5')
            and not url.endswith('.png')
            and url != self.properties['url']
            and 's3credentials' not in url
        ]

    def find_urls(self, extension: str = None, pattern: str = r'.*', directAccess: bool = False) -> List[str]:
        """
        Searches for all urls matching a given extension and/or pattern
        param extension: the file extension to search for. (Defaults to `None`)
            - Example: '.tiff'
        param pattern: A regex pattern to search each url for.(Defaults to `False`)
            - Example: `r'(QA_)+'` to find urls with 'QA_' at least once
        param directAccess: should search in s3 bucket urls (Defaults to `False`)
        """
        search_list = self._get_s3_uris() if directAccess else self._get_urls()
        
        def _get_extension(file_url: str):
            path = parse.urlparse(file_url).path
            return os.path.splitext(path)[-1]
        
        if extension is not None:
            search_list = [url for url in search_list if _get_extension(url) == extension] 

        regexp = re.compile(pattern=pattern)

        return sorted([url for url in search_list if regexp.search(url) is not None])

    def centroid(self) -> Point:
        """
        Finds the centroid of a product
        """
        coords = mapping(shape(self.geometry))['coordinates'][0]
        lons = [p[0] for p in coords]
        if max(lons) - min(lons) > 180:
            unwrapped_coords = [a if a[0] > 0 else [a[0] + 360, a[1]] for a in coords]
        else:
            unwrapped_coords = [a for a in coords]

        return Polygon(unwrapped_coords).centroid

    def remotezip(self, session: ASFSession) -> 'RemoteZip':  # type: ignore # noqa: F821
        """Returns a RemoteZip object which can be used to download
        a part of an ASFProduct's zip archive. (See example in examples/5-Download.ipynb)

        requires installing optional dependencies via pip or conda to use the `remotezip` package:

        `python3 -m pip install asf-search[extras]`

        :param session: an authenticated ASFSession
        """
        from .download.download import remotezip

        return remotezip(self.properties['url'], session=session)

    def _read_umm_property(self, umm: Dict, mapping: Dict) -> Any:
        value = self.umm_get(umm, *mapping['path'])
        if mapping.get('cast') is None:
            return value

        return self.umm_cast(mapping['cast'], value)

    def _get_geometry(self, item: Dict):
        """Helper method that creates a geometry context dictionary. 
        Meant primarily for NISARProduct to override for dateline multipolygon parsing.
        """
        try:
            coordinates = item['umm']['SpatialExtent']['HorizontalSpatialDomain']['Geometry'][
                'GPolygons'
            ][0]['Boundary']['Points']
            coordinates = [[c['Longitude'], c['Latitude']] for c in coordinates]
            geometry = {'coordinates': [coordinates], 'type': 'Polygon'}
        except KeyError:
            geometry = {'coordinates': None, 'type': 'Polygon'}

        return geometry
    
    def translate_product(self, item: Dict) -> Dict:
        """
        Generates `properties` and `geometry` from the CMR UMM response
        """
        geometry = self._get_geometry(item)
        umm = item.get('umm')

        # additionalAttributes = {attr['Name']: attr['Values'] for attr in umm['AdditionalAttributes']}

        properties = {
            prop: self._read_umm_property(umm, umm_mapping)
            for prop, umm_mapping in self._base_properties.items()
        }

        if properties.get('url') is not None:
            properties['fileName'] = properties['url'].split('/')[-1]
        else:
            properties['fileName'] = None

        # Fallbacks
        if properties.get('beamModeType') is None:
            properties['beamModeType'] = self.umm_get(
                umm, 'AdditionalAttributes', ('Name', 'BEAM_MODE'), 'Values', 0
            )

        if properties.get('platform') is None:
            properties['platform'] = self.umm_get(umm, 'Platforms', 0, 'ShortName')

        return {'geometry': geometry, 'properties': properties, 'type': 'Feature'}

    def get_sort_keys(self) -> Tuple[str, str]:
        """
        Returns tuple of primary and secondary date values used for sorting final search results
        Any subclasses must return string for final `sort()` to work
        """
        # `sort()` will raise an error when comparing `NoneType`,
        # using self._read_property() to wrap standard `dict.get()` for possible `None` values
        primary_key = self._read_property(key='stopTime', default='')
        secondary_key = self._read_property(
            key='fileID', default=self._read_property('sceneName', '')
        )

        return (primary_key, secondary_key)

    def _read_property(self, key: str, default: Any = None) -> Any:
        """
        Helper method wraps `properties.get()`.
        Since a property can be `None`, if the key exists`dict.get('key', 'default')`
        will never return the default
        """
        output = default
        if (value := self.properties.get(key)) is not None:
            output = value

        return output

    @final
    @staticmethod
    def umm_get(item: Dict, *args):
        """
        Used to search for values in CMR UMM

        :param item: the umm dict returned from CMR
        :param *args: the expected path to the value

        Example case:
        "I want to grab the polarization from the granule umm"
        ```
        item = {
            'AdditionalAttributes': [
                {
                    'Name': 'POLARIZATION',
                    'Values': ['VV', 'VH']
                },
                ...
            ],
            ...
        }
        ```

        The path provided to *args would look like this:
        ```
        'AdditionalAttributes', ('Name', 'POLARIZATION'), 'Values', 0
        result: 'VV'
        ```

        - `'AdditionalAttributes'` acts like item['AdditionalAttributes'],
        which is a list of dictionaries

        - Since `AdditionalAttributes` is a LIST of dictionaries,
        we search for a dict with the key value pair,
        `('Name', 'POLARIZATION')`

        - If found, we try to access that dictionary's `Values` key
        - Since `Values` is a list, we can access the first index `0` (in this case, 'VV')

        ---

        If you want more of the umm, simply reduce how deep you search:
        Example: "I need BOTH polarizations (`OPERAS1Product` does this, noticed the omitted `0`)

        ```
        'AdditionalAttributes', ('Name', 'POLARIZATION'), 'Values'
        result: ['VV', 'VH']
        ```

        ---

        Example: "I need the ENTIRE POLARIZATION dict"

        ```
        'AdditionalAttributes', ('Name', 'POLARIZATION')
        result: {
                    'Name': 'POLARIZATION',
                    'Values': ['VV', 'VH']
                }
        ```

        ---

        ADVANCED:
        Sometimes there are multiple dictionaries in a list that have
        the same key value pair we're searching for (See `OPERAS1Product` umm under `RelatedUrls`).
        This means we can miss values since we're only grabbing the first match
        depending on how the umm is organized.
        There is a way to get ALL data that matches our key value criteria.

        Example: "I need ALL `URL` values for dictionaries in `RelatedUrls`
        where `Type` is `GET DATA`" (See in use in `OPERAS1Product` class)
        ```
        'RelatedUrls', ('Type', [('GET DATA', 'URL')]), 0
        ```
        """
        if item is None:
            return None
        for key in args:
            if isinstance(key, str):
                item = item.get(key)
            elif isinstance(key, int):
                item = item[key] if key < len(item) else None
            elif isinstance(key, tuple):
                (a, b) = key
                if isinstance(b, List):
                    output = []
                    b = b[0]
                    for child in item:
                        if ASFProduct.umm_get(child, key[0]) == b[0]:
                            output.append(ASFProduct.umm_get(child, b[1]))
                    if len(output):
                        return output

                    return None

                found = False
                for child in item:
                    if ASFProduct.umm_get(child, a) == b:
                        item = child
                        found = True
                        break
                if not found:
                    return None
            if item is None:
                return None
        if item in [None, 'NA', 'N/A', '']:
            item = None
        return item

    @final
    @staticmethod
    def umm_cast(f, v):
        """Tries to cast value v by callable f, returns None if it fails"""
        try:
            return f(v)
        except TypeError:
            return None

    @staticmethod
    def _is_subclass(item: Dict) -> bool:
        """
        Used to determine which subclass to use for specific
        edge-cases when parsing results in search methods
        (Currently implemented for ARIA and OPERA subclasses).

        params:
        - item (dict): the CMR UMM-G item to read from
        """
        raise NotImplementedError()



from collections import UserList
from multiprocessing import Pool
import json
from typing import List
from asf_search import ASFSession, ASFSearchOptions
from asf_search.download.file_download_type import FileDownloadType
from asf_search.exceptions import ASFSearchError

from asf_search import ASF_LOGGER
from asf_search.export.csv import results_to_csv
from asf_search.export.jsonlite import results_to_jsonlite
from asf_search.export.jsonlite2 import results_to_jsonlite2
from asf_search.export.json import results_to_json
from asf_search.export.kml import results_to_kml
from asf_search.export.metalink import results_to_metalink


class ASFSearchResults(UserList):
    def __init__(self, *args, opts: ASFSearchOptions = None):
        super().__init__(*args)
        # Store it JUST so the user can access it (There might be zero products)
        # Each product will use their own reference to opts (but points to the same obj)
        self.searchOptions = opts
        self.searchComplete = False

    def geojson(self):
        return {
            'type': 'FeatureCollection',
            'features': [product.geojson() for product in self],
        }

    def csv(self):
        return results_to_csv(self)

    def kml(self):
        return results_to_kml(self)

    def metalink(self):
        return results_to_metalink(self)

    def json(self):
        return results_to_json(self)

    def jsonlite(self):
        return results_to_jsonlite(self)

    def jsonlite2(self):
        return results_to_jsonlite2(self)

    def find_urls(self, extension: str = None, pattern: str = r'.*', directAccess: bool = False) -> List[str]:
        """Returns a flat list of all https or s3 urls from all results matching an extension and/or regex pattern
        param extension: the file extension to search for. (Defaults to `None`)
            - Example: '.tiff'
        param pattern: A regex pattern to search each url for.(Defaults to `False`)
            - Example: `r'(QA_)+'` to find urls with 'QA_' at least once
        param directAccess: should search in s3 bucket urls (Defaults to `False`)
        """
        urls = []

        for product in self:
            urls.extend(product.find_urls(extension=extension, pattern=pattern, directAccess=directAccess))
        
        return sorted(list(set(urls)))
    
    def __str__(self):
        return json.dumps(self.geojson(), indent=2, sort_keys=True)

    def download(
        self,
        path: str,
        session: ASFSession = None,
        processes: int = 1,
        fileType=FileDownloadType.DEFAULT_FILE,
    ) -> None:
        """
        Iterates over each ASFProduct and downloads them to the specified path.

        Parameters
        ----------
        path:
            The directory into which the products should be downloaded.
        session:
            The session to use
            Defaults to the session used to fetch the results, or a new one if none was used.
        processes:
            Number of download processes to use. Defaults to 1 (i.e. sequential download)

        """
        ASF_LOGGER.info(f'Started downloading ASFSearchResults of size {len(self)}.')
        if processes == 1:
            for product in self:
                product.download(path=path, session=session, fileType=fileType)
        else:
            ASF_LOGGER.info(f'Using {processes} threads - starting up pool.')
            pool = Pool(processes=processes)
            args = [(product, path, session, fileType) for product in self]
            pool.map(_download_product, args)
            pool.close()
            pool.join()
        ASF_LOGGER.info(f'Finished downloading ASFSearchResults of size {len(self)}.')

    def raise_if_incomplete(self) -> None:
        if not self.searchComplete:
            msg = (
                'Results are incomplete due to a search error. '
                'See logging for more details. (ASFSearchResults.raise_if_incomplete called)'
            )

            ASF_LOGGER.error(msg)
            raise ASFSearchError(msg)

    def get_products_by_subclass_type(self) -> dict:
        """
        Organizes results into dictionary by ASFProduct subclass name
        : return: Dict of ASFSearchResults, organized by ASFProduct subclass names
        """
        subclasses = {}

        for product in self.data:
            product_type = product.get_classname()

            if subclasses.get(product_type) is None:
                subclasses[product_type] = ASFSearchResults([])

            subclasses[product_type].append(product)

        return subclasses


def _download_product(args) -> None:
    product, path, session, fileType = args
    product.download(path=path, session=session, fileType=fileType)



import numbers
from typing import List
import asf_search
from asf_search.ASFSession import ASFSession
from requests.cookies import create_cookie
import http.cookiejar
import requests
from multiprocessing import Pool

from unittest.mock import patch


def run_auth_with_creds(username: str, password: str):
    session = ASFSession()
    session.auth_with_creds(username=username, password=password)


def run_auth_with_token(token: str):
    session = ASFSession()

    with patch('asf_search.ASFSession.post') as mock_token_session:
        if not token.startswith('Bearer EDL'):
            mock_token_session.return_value.status_code = 400
            session.auth_with_token(token)

        mock_token_session.return_value.status_code = 200
        session.auth_with_token(token)


def run_auth_with_cookiejar(cookies: List):
    cookiejar = http.cookiejar.CookieJar()
    for cookie in cookies:
        cookiejar.set_cookie(create_cookie(name=cookie.pop('name'), **cookie))

    # requests.cookies.RequestsCookieJar, which has slightly different behaviour
    session = ASFSession()
    session.auth_with_cookiejar(cookiejar)

    request_cookiejar_session = ASFSession()
    request_cookiejar_session.auth_with_cookiejar(session.cookies)


def run_test_asf_session_rebuild_auth(
    original_domain: str,
    response_domain: str,
    response_code: numbers.Number,
    final_token,
):
    if final_token == 'None':
        final_token = None

    session = ASFSession()

    with patch('asf_search.ASFSession.post') as mock_token_session:
        mock_token_session.return_value.status_code = 200
        session.auth_with_token('bad_token')

        req = requests.Request(original_domain)
        req.headers.update({'Authorization': 'Bearer fakeToken'})

        response = requests.Response()
        response.status_code = response_code
        response.location = response_domain
        response.request = requests.Request()
        response.request.url = response_domain
        response.headers.update({'Authorization': 'Bearer fakeToken'})

        with patch('asf_search.ASFSession._get_domain') as hostname_patch:
            hostname_patch.side_effect = [original_domain, response_domain]

            session.rebuild_auth(req, response)

            assert req.headers.get('Authorization') == final_token


def test_ASFSession_INTERNAL_mangling():
    session = asf_search.ASFSession()
    session.cmr_host = asf_search.constants.INTERNAL.EDL_HOST
    session.edl_host = asf_search.constants.INTERNAL.EDL_CLIENT_ID
    session.auth_cookie_names = asf_search.constants.INTERNAL.ASF_AUTH_HOST
    session.auth_domains = asf_search.constants.INTERNAL.CMR_HOST
    session.asf_auth_host = asf_search.constants.INTERNAL.CMR_COLLECTIONS
    session.cmr_collections = asf_search.constants.INTERNAL.AUTH_DOMAINS
    session.edl_client_id = asf_search.constants.INTERNAL.AUTH_COOKIES

    # get the current defaults since we're going to mangle them
    DEFAULT_EDL_HOST = asf_search.constants.INTERNAL.EDL_HOST
    DEFAULT_EDL_CLIENT_ID = asf_search.constants.INTERNAL.EDL_CLIENT_ID
    DEFAULT_ASF_AUTH_HOST = asf_search.constants.INTERNAL.ASF_AUTH_HOST
    DEFAULT_CMR_HOST = asf_search.constants.INTERNAL.CMR_HOST
    DEFAULT_CMR_COLLECTIONS = asf_search.constants.INTERNAL.CMR_COLLECTIONS
    DEFAULT_AUTH_DOMAINS = asf_search.constants.INTERNAL.AUTH_DOMAINS
    DEFAULT_AUTH_COOKIES = asf_search.constants.INTERNAL.AUTH_COOKIES

    uat_domain = 'cmr.uat.earthdata.nasa.gov'
    edl_client_id = 'custom_client_id'
    auth_host = 'custom_auth_host'
    cmr_collection = '/search/granules'
    auth_domains = ['custom_auth_domain']
    uat_login_cookie = ['uat_urs_user_already_logged']
    uat_login_domain = 'uat.urs.earthdata.nasa.gov'

    asf_search.constants.INTERNAL.CMR_HOST = uat_domain
    asf_search.constants.INTERNAL.EDL_HOST = uat_login_domain
    asf_search.constants.INTERNAL.AUTH_COOKIES = uat_login_cookie
    asf_search.constants.INTERNAL.EDL_CLIENT_ID = edl_client_id
    asf_search.constants.INTERNAL.AUTH_DOMAINS = auth_domains
    asf_search.constants.INTERNAL.ASF_AUTH_HOST = auth_host
    asf_search.constants.INTERNAL.CMR_COLLECTIONS = cmr_collection

    mangeled_session = asf_search.ASFSession()

    # set them back
    asf_search.constants.INTERNAL.EDL_HOST = DEFAULT_EDL_HOST
    asf_search.constants.INTERNAL.EDL_CLIENT_ID = DEFAULT_EDL_CLIENT_ID
    asf_search.constants.INTERNAL.ASF_AUTH_HOST = DEFAULT_ASF_AUTH_HOST
    asf_search.constants.INTERNAL.CMR_HOST = DEFAULT_CMR_HOST
    asf_search.constants.INTERNAL.CMR_COLLECTIONS = DEFAULT_CMR_COLLECTIONS
    asf_search.constants.INTERNAL.AUTH_DOMAINS = DEFAULT_AUTH_DOMAINS
    asf_search.constants.INTERNAL.AUTH_COOKIES = DEFAULT_AUTH_COOKIES

    assert mangeled_session.cmr_host == uat_domain
    assert mangeled_session.edl_host == uat_login_domain
    assert mangeled_session.auth_cookie_names == uat_login_cookie
    assert mangeled_session.auth_domains == auth_domains
    assert mangeled_session.asf_auth_host == auth_host
    assert mangeled_session.cmr_collections == cmr_collection
    assert mangeled_session.edl_client_id == edl_client_id

    custom_session = asf_search.ASFSession(
        cmr_host=uat_domain,
        edl_host=uat_login_domain,
        auth_cookie_names=uat_login_cookie,
        auth_domains=auth_domains,
        asf_auth_host=auth_host,
        cmr_collections=cmr_collection,
        edl_client_id=edl_client_id,
    )

    assert custom_session.cmr_host == uat_domain
    assert custom_session.edl_host == uat_login_domain
    assert custom_session.auth_cookie_names == uat_login_cookie
    assert custom_session.auth_domains == auth_domains
    assert custom_session.asf_auth_host == auth_host
    assert custom_session.cmr_collections == cmr_collection
    assert custom_session.edl_client_id == edl_client_id


def test_ASFSession_pooling():
    uat_domain = 'cmr.uat.earthdata.nasa.gov'
    edl_client_id = 'custom_client_id'
    auth_host = 'custom_auth_host'
    cmr_collection = '/search/granules'
    auth_domains = ['custom_auth_domain']
    uat_login_cookie = ['uat_urs_user_already_logged']
    uat_login_domain = 'uat.urs.earthdata.nasa.gov'

    custom_session = asf_search.ASFSession(
        cmr_host=uat_domain,
        edl_host=uat_login_domain,
        auth_cookie_names=uat_login_cookie,
        auth_domains=auth_domains,
        asf_auth_host=auth_host,
        cmr_collections=cmr_collection,
        edl_client_id=edl_client_id,
    )
    Pool()
    pool = Pool(processes=2)
    pool.map(_assert_pooled_instance_variables, [custom_session, custom_session])
    pool.close()
    pool.join()


def _assert_pooled_instance_variables(session):
    uat_domain = 'cmr.uat.earthdata.nasa.gov'
    edl_client_id = 'custom_client_id'
    auth_host = 'custom_auth_host'
    cmr_collection = '/search/granules'
    auth_domains = ['custom_auth_domain']
    uat_login_cookie = ['uat_urs_user_already_logged']
    uat_login_domain = 'uat.urs.earthdata.nasa.gov'
    assert session.cmr_host == uat_domain
    assert session.edl_host == uat_login_domain
    assert session.auth_cookie_names == uat_login_cookie
    assert session.auth_domains == auth_domains
    assert session.asf_auth_host == auth_host
    assert session.cmr_collections == cmr_collection
    assert session.edl_client_id == edl_client_id



from .MissionList import get_campaigns  # noqa: F401
from .subquery import build_subqueries  # noqa: F401
from .translate import translate_opts  # noqa: F401
from .field_map import field_map  # noqa: F401
from .datasets import (  # noqa: F401
    dataset_collections,  # noqa: F401
    collections_per_platform,  # noqa: F401
    collections_by_processing_level,  # noqa: F401
    get_concept_id_alias,  # noqa: F401
    get_dataset_concept_ids,  # noqa: F401
)



import inspect
import json
import re
from types import GeneratorType
from typing import Tuple
from shapely.geometry import shape
from shapely.ops import transform

from asf_search import ASF_LOGGER
from asf_search.export.export_translators import ASFSearchResults_to_properties_list

extra_json_fields = [
    (
        'processingTypeDisplay',
        ['AdditionalAttributes', ('Name', 'PROCESSING_TYPE_DISPLAY'), 'Values', 0],
    ),
    (
        'beamMode',
        ['AdditionalAttributes', ('Name', 'BEAM_MODE'), 'Values', 0],
    ),
    ('thumbnailUrl', ['AdditionalAttributes', ('Name', 'THUMBNAIL_URL'), 'Values', 0]),
    (
        'faradayRotation',
        ['AdditionalAttributes', ('Name', 'FARADAY_ROTATION'), 'Values', 0],
    ),
    ('sizeMB', ['DataGranule', 'ArchiveAndDistributionInformation', 0, 'Size']),
    ('flightLine', ['AdditionalAttributes', ('Name', 'FLIGHT_LINE'), 'Values', 0]),
    ('missionName', ['AdditionalAttributes', ('Name', 'MISSION_NAME'), 'Values', 0]),
    ('centerLat', ['AdditionalAttributes', ('Name', 'CENTER_LAT'), 'Values', 0]),
    ('centerLon', ['AdditionalAttributes', ('Name', 'CENTER_LON'), 'Values', 0]),
    ('configurationName', ['AdditionalAttributes', ('Name', 'BEAM_MODE_DESC'), 'Values', 0]),
    ('doppler', ['AdditionalAttributes', ('Name', 'DOPPLER'), 'Values', 0]),
    ('nearStartLat', ['AdditionalAttributes', ('Name', 'NEAR_START_LAT'), 'Values', 0]),
    ('nearStartLon', ['AdditionalAttributes', ('Name', 'NEAR_START_LON'), 'Values', 0]),
    ('farStartLat', ['AdditionalAttributes', ('Name', 'FAR_START_LAT'), 'Values', 0]),
    ('farStartLon', ['AdditionalAttributes', ('Name', 'FAR_START_LON'), 'Values', 0]),
    ('nearEndLat', ['AdditionalAttributes', ('Name', 'NEAR_END_LAT'), 'Values', 0]),
    ('nearEndLon', ['AdditionalAttributes', ('Name', 'NEAR_END_LON'), 'Values', 0]),
    ('farEndLat', ['AdditionalAttributes', ('Name', 'FAR_END_LAT'), 'Values', 0]),
    ('farEndLon', ['AdditionalAttributes', ('Name', 'FAR_END_LON'), 'Values', 0]),
    (
        'faradayRotation',
        ['AdditionalAttributes', ('Name', 'FARADAY_ROTATION'), 'Values', 0],
    ),
    (
        'finalFrame',
        ['AdditionalAttributes', ('Name', 'CENTER_ESA_FRAME'), 'Values', 0],
    ),
    (
        'firstFrame',
        ['AdditionalAttributes', ('Name', 'CENTER_ESA_FRAME'), 'Values', 0],
    ),
    (
        'insarStackSize',
        ['AdditionalAttributes', ('Name', 'INSAR_STACK_SIZE'), 'Values', 0],
    ),
    (
        'lookDirection',
        ['AdditionalAttributes', ('Name', 'LOOK_DIRECTION'), 'Values', 0],
    ),
    (
        'processingDescription',
        ['AdditionalAttributes', ('Name', 'PROCESSING_DESCRIPTION'), 'Values', 0],
    ),
    ('sceneDate', ['AdditionalAttributes', ('Name', 'ACQUISITION_DATE'), 'Values', 0]),
    ('processingType', ['AdditionalAttributes', ('Name', 'PROCESSING_LEVEL'), 'Values', 0]),
    ('pointingAngle', ['AdditionalAttributes', ('Name', 'POINTING_ANGLE'), 'Values', 0]),
    ('offNadirAngle', ['AdditionalAttributes', ('Name', 'OFF_NADIR_ANGLE'), 'Values', 0]),
]


def results_to_json(results):
    ASF_LOGGER.info('started translating results to json format')
    if len(results) == 0:
        yield from json.JSONEncoder(indent=2, sort_keys=True).iterencode([[]])
        return

    if not inspect.isgeneratorfunction(results) and not isinstance(results, GeneratorType):
        results = [results]

    streamer = JsonStreamArray(results)

    for p in json.JSONEncoder(indent=2, sort_keys=True).iterencode([streamer]):
        yield p


def get_wkts(geometry) -> Tuple[str, str]:
    wrapped = shape(geometry)

    min_lon, max_lon = (wrapped.bounds[0], wrapped.bounds[2])

    if max_lon - min_lon > 180:
        unwrapped = transform(unwrap_shape, wrapped)
    else:
        unwrapped = wrapped

    return wrapped.wkt, unwrapped.wkt


class JsonStreamArray(list):
    def __init__(self, results):
        self.results = results

        # need to make sure we actually have results so we can intelligently set __len__, otherwise
        # iterencode behaves strangely and will output invalid json
        self.len = 1

    def __iter__(self):
        return self.streamDicts()

    def __len__(self):
        return self.len

    def get_additional_output_fields(self, product):
        # umm = product.umm

        additional_fields = {}
        for key, path in extra_json_fields:
            additional_fields[key] = product.umm_get(product.umm, *path)

        platform = product.properties.get('platform')
        if platform is None:
            platform = ''
        if platform.upper() in [
            'ALOS',
            'RADARSAT-1',
            'JERS-1',
            'ERS-1',
            'ERS-2',
        ]:
            insarGrouping = product.umm_get(
                product.umm,
                *['AdditionalAttributes', ('Name', 'INSAR_STACK_ID'), 'Values', 0],
            )

            if insarGrouping not in [None, 0, '0', 'NA', 'NULL']:
                additional_fields['canInsar'] = True
                additional_fields['insarStackSize'] = product.umm_get(
                    product.umm,
                    *[
                        'AdditionalAttributes',
                        ('Name', 'INSAR_STACK_SIZE'),
                        'Values',
                        0,
                    ],
                )
            else:
                additional_fields['canInsar'] = False
        else:
            additional_fields['canInsar'] = product.baseline is not None

        additional_fields['geometry'] = product.geometry

        return additional_fields

    def streamDicts(self):
        completed = False
        for page_idx, page in enumerate(self.results):
            ASF_LOGGER.info(f'Streaming {len(page)} products from page {page_idx}')
            completed = page.searchComplete

            yield from [
                self.getItem(p)
                for p in ASFSearchResults_to_properties_list(
                    page, self.get_additional_output_fields
                )
                if p is not None
            ]

        if not completed:
            ASF_LOGGER.warn('Failed to download all results from CMR')

        ASF_LOGGER.info(f'Finished streaming {self.getOutputType()} results')

    def getItem(self, p):
        for i in p.keys():
            if p[i] == 'NA' or p[i] == '':
                p[i] = None
        try:
            if p.get('offNadirAngle') is not None and float(p['offNadirAngle']) < 0:
                p['offNadirAngle'] = None
        except TypeError:
            pass

        try:
            if p.get('patNumber'):
                if float(p['pathNumber']) < 0:
                    p['pathNumber'] = None
        except TypeError:
            pass

        try:
            if p.get('groupID') is None:
                p['groupID'] = p['sceneName']
        except TypeError:
            pass

        try:
            p['sizeMB'] = float(p['sizeMB'])
        except TypeError:
            pass

        try:
            p['pathNumber'] = int(p['pathNumber'])
        except TypeError:
            pass

        try:
            p['frameNumber'] = int(p.get('frameNumber'))
        except TypeError:
            pass

        try:
            p['orbit'] = int(p['orbit'])
        except TypeError:
            pass

        wrapped, unwrapped = get_wkts(p['geometry'])

        result = {
            'absoluteOrbit': p.get('orbit'),  #
            'beamMode': p.get('beamMode'),
            'beamModeType': p.get('beamModeType'),  #
            'browse': p.get('browse'),
            'centerLat': p.get('centerLat'),
            'centerLon': p.get('centerLon'),
            'collectionName': p.get('missionName'),
            'configurationName': p.get('configurationName'),  #
            'doppler': p.get('doppler'),  #
            'downloadUrl': p.get('url'),
            'farEndLat': p.get('farEndLat'),
            'farEndLon': p.get('farEndLon'),
            'farStartLat': p.get('farStartLat'),
            'farStartLon': p.get('farStartLon'),
            'faradayRotation': p.get('faradayRotation'),
            'fileName': p.get('fileName'),
            'finalFrame': p.get('finalFrame'),
            'firstFrame': p.get('firstFrame'),
            'flightDirection': p.get('flightDirection'),
            'flightLine': p.get('flightLine'),
            'frameNumber': p.get('frameNumber'),
            'granuleName': p.get('sceneName'),
            'granuleType': p.get('granuleType'),
            'groupID': p.get('groupID'),
            'insarGrouping': p.get('insarGrouping'),
            'insarStackSize': p.get('insarStackSize'),
            'lookDirection': p.get('lookDirection'),
            'md5sum': p.get('md5sum'),
            'missionName': p.get('missionName'),
            'nearEndLat': p.get('nearEndLat'),
            'nearEndLon': p.get('nearEndLon'),
            'nearStartLat': p.get('nearStartLat'),
            'nearStartLon': p.get('nearStartLon'),  #
            'offNadirAngle': p.get('offNadirAngle'),
            'platform': p.get('platform'),
            'pointingAngle': p.get('pointingAngle'),  #
            'polarization': p.get('polarization'),
            'processingDate': p.get('processingDate'),
            'processingDescription': p.get('processingDescription'),
            'processingLevel': p.get('processingLevel'),
            'processingType': p.get('processingType'),
            'processingTypeDisplay': p.get('processingTypeDisplay'),  #
            'productName': p.get('sceneName'),
            'product_file_id': p.get('fileID'),
            'relativeOrbit': p.get('pathNumber'),
            'sceneDate': p.get('sceneDate'),  #
            'sceneId': p.get('sceneName'),  #
            'sensor': p.get('sensor'),
            'sizeMB': p.get('sizeMB'),
            'startTime': p.get('startTime'),
            'stopTime': p.get('stopTime'),
            'stringFootprint': wrapped,
            'thumbnailUrl': p.get('thumbnailUrl'),
            'track': p.get('pathNumber'),
            'subswath': p.get('subswath'),
            'pgeVersion': p.get('pgeVersion'),
        }

        for key in result.keys():
            if result[key] in ['NA', 'NULL']:
                result[key] = None

        if 'temporalBaseline' in p.keys():
            result['temporalBaseline'] = p['temporalBaseline']
        if 'perpendicularBaseline' in p.keys():
            result['perpendicularBaseline'] = p['perpendicularBaseline']

        if p.get('processingLevel') == 'BURST':  # is a burst product
            result['burst'] = p['burst']
            result['sizeMB'] = float(p['bytes']) / 1024000

        elif p.get('operaBurstID') is not None or p.get('fileID').startswith('OPERA'):
            result['opera'] = {
                'operaBurstID': p.get('operaBurstID'),
                's3Urls': p.get('s3Urls', []),
                'additionalUrls': p.get('additionalUrls'),
            }
            if p.get('validityStartDate'):
                result['opera']['validityStartDate'] = p.get('validityStartDate')
        elif p.get('platform') == 'NISAR':
            result['nisar'] = {
                'additionalUrls': p.get('additionalUrls', []),
                's3Urls': p.get('s3Urls', []),
                'pgeVersion': p.get('pgeVersion'),
                'mainBandPolarization': p.get('mainBandPolarization'),
                'sideBandPolarization': p.get('sideBandPolarization'),
                'frameCoverage': p.get('frameCoverage'),
                'jointObservation': p.get('jointObservation'),
                'rangeBandwidth': p.get('rangeBandwidth'),
            }
        elif result.get('productID', result.get('fileName', '')).startswith('S1-GUNW'):
            result.pop('perpendicularBaseline', None)
            if p.get('ariaVersion') is None:
                version_unformatted = result.get('productID', '').split('v')[-1]
                result['ariaVersion'] = re.sub(
                    r'[^0-9\.]', '', version_unformatted.replace('_', '.')
                )
            else:
                result['ariaVersion'] = p.get('ariaVersion')

        if p.get('browse') is not None and len(p['browse']) == 1:
            result['browse'] = p['browse'][0]

        return result

    def getOutputType(self) -> str:
        return 'json'


def unwrap_shape(x, y, z=None):
    x = x if x > 0 else x + 360
    return tuple([x, y])



class ASFWarning(Warning):
    """
    Base ASF Warning, not intended for direct use

    Tip: Silence me to silence all child ASFWarnings
    """

class PairNotInFullStackWarning(ASFWarning):
    """Warn when attempting to do something with a Pair that is not in Stack.full_stack"""



from asf_search import ASF_LOGGER

from .validators import (
    parse_string,
    parse_float,
    parse_int,
    parse_wkt,
    parse_date,
    parse_string_list,
    parse_int_list,
    parse_int_or_range_list,
    parse_float_or_range_list,
    parse_cmr_keywords_list,
    parse_session,
    parse_circle,
    parse_linestring,
    parse_point,
    parse_bbox,
)


def validate(key, value):
    if key not in validator_map:
        error_msg = f'Key "{key}" is not a valid search option.'
        # See if they just missed up case sensitivity:
        for valid_key in validator_map:
            if key.lower() == valid_key.lower():
                error_msg += f' (Did you mean "{valid_key}"?)'
                break
        ASF_LOGGER.error(error_msg)
        raise KeyError(error_msg)
    try:
        return validator_map[key](value)
    except ValueError as exc:
        ASF_LOGGER.exception(f'Failed to parse item in ASFSearchOptions: {key=} {value=} {exc=}')
        raise


validator_map = {
    # Search parameters       Parser
    'maxResults': int,
    'absoluteOrbit': parse_int_or_range_list,
    'asfFrame': parse_int_or_range_list,
    'bbox': parse_bbox,
    'beamMode': parse_string_list,
    'beamSwath': parse_string_list,
    'campaign': parse_string,
    'circle': parse_circle,
    'linestring': parse_linestring,
    'point': parse_point,
    'maxDoppler': parse_float,
    'minDoppler': parse_float,
    'maxBaselinePerp': parse_float,
    'minBaselinePerp': parse_float,
    'maxInsarStackSize': parse_int,
    'minInsarStackSize': parse_int,
    'maxFaradayRotation': parse_float,
    'minFaradayRotation': parse_float,
    'flightDirection': parse_string,
    'flightLine': parse_string,
    'frame': parse_int_or_range_list,
    'granule_list': parse_string_list,
    'product_list': parse_string_list,
    'intersectsWith': parse_wkt,
    'lookDirection': parse_string,
    'offNadirAngle': parse_float_or_range_list,
    'platform': parse_string_list,
    'polarization': parse_string_list,
    'processingLevel': parse_string_list,
    'relativeOrbit': parse_int_or_range_list,
    'processingDate': parse_date,
    'start': parse_date,
    'end': parse_date,
    'season': parse_int_list,
    'groupID': parse_string_list,
    'insarStackId': parse_string,
    'instrument': parse_string_list,
    'collections': parse_string_list,
    'shortName': parse_string_list,
    'dataset': parse_string_list,
    'cmr_keywords': parse_cmr_keywords_list,
    'ariaVersion': parse_string,
    'tileID': parse_string_list,
    # S1 Inteferrogram Filters
    'temporalBaselineDays': parse_string_list,
    # Opera Burst Filters
    'operaBurstID': parse_string_list,
    # SLC Burst Filters
    'absoluteBurstID': parse_int_list,
    'relativeBurstID': parse_int_list,
    'fullBurstID': parse_string_list,
    # nisar paramaters
    'frameCoverage': parse_string,
    'jointObservation': bool,
    'mainBandPolarization': parse_string_list,
    'sideBandPolarization': parse_string_list,
    'rangeBandwidth': parse_string_list,
    'productionConfiguration': parse_string_list,
    # Config parameters       Parser
    'session': parse_session,
    'host': parse_string,
    'provider': parse_string,
    'collectionAlias': bool,
}



from .search import search  # noqa: F401
from .granule_search import granule_search  # noqa: F401
from .product_search import product_search  # noqa: F401
from .geo_search import geo_search  # noqa: F401
from .baseline_search import stack_from_id  # noqa: F401
from .campaigns import campaigns  # noqa: F401
from .search_count import search_count  # noqa: F401
from .search_generator import search_generator, preprocess_opts  # noqa: F401
from .collection_attributes import get_searchable_attributes  # noqa: F401



import pickle
from pathlib import Path

from asf_search.ASFSearchOptions import ASFSearchOptions
from asf_search import Stack
from asf_search.warnings import PairNotInFullStackWarning

import pytest

@pytest.fixture
def reference():
    path = Path(__file__).parent / "data/reference_scene_0.pkl"

    with open(path, "rb") as f:
        # trusted local test fixture
        return pickle.load(f) # nosec B301
    
@pytest.fixture
def pair():
    path = Path(__file__).parent / "data/pair_0.pkl"

    with open(path, "rb") as f:
        # trusted local test fixture
        return pickle.load(f) # nosec B301
    
@pytest.fixture
def stack_results():
    path = Path(__file__).parent / "data/stack_search_0.pkl"

    with open(path, "rb") as f:
        # trusted local test fixture
        return pickle.load(f) # nosec B301
    
@pytest.fixture
def stack(stack_results):
    return Stack.from_search_results(stack_results)
    
def test_stack_from_ref_scene(mocker, reference, stack_results):
    """
    Create a Stack from a geographic reference scene and confrm expected size of its full_stack and subset_stack
    """
    args = ASFSearchOptions(
        **{"start": "2022-01-01", "end": "2022-04-02"}
    )

    mock_stack = mocker.patch.object(
        reference,
        "stack",
        return_value=stack_results,
    )

    stack = Stack(reference, opts=args)

    mock_stack.assert_called_once_with(opts=args)

    assert len(stack.full_stack) == 21
    assert stack.subset_stack == stack.full_stack

def test_stack_from_search_results(stack):
    """
    Create a Stack from ASFProduct.stack search results with the Stack.from_search_results alternate class method constructor
    """
    assert len(stack.full_stack) == 21
    assert stack.subset_stack == stack.full_stack

def test_remove_pair(stack):
    stack.remove_pairs(stack.full_stack[0])
    assert len(stack.subset_stack) == 20
    assert len(stack.remove_list) == 1
    assert len(stack.connected_substacks) == 1 

def test_remove_pairs(stack):
    stack.remove_pairs(stack.full_stack[1:11])
    assert len(stack.subset_stack) == 11
    assert len(stack.remove_list) == 10
    assert len(stack.connected_substacks) == 2

def test_attempt_remove_invalid_pair(stack, pair):
    with pytest.warns(PairNotInFullStackWarning):
        stack.remove_pairs(pair)

def test_add_pair(stack):
    stack.remove_pairs(stack.full_stack[0])
    stack.add_pairs(stack.remove_list[0])
    assert len(stack.subset_stack) == 21
    assert len(stack.remove_list) == 0
    assert len(stack.connected_substacks) == 1 

def test_add_pairs(stack):
    stack.remove_pairs(stack.full_stack[1:11])
    stack.add_pairs(stack.remove_list)
    assert len(stack.subset_stack) == 21
    assert len(stack.remove_list) == 0
    assert len(stack.connected_substacks) == 1

def test_add_new_pair(stack, pair):
    stack.add_pairs(pair)
    assert len(stack.subset_stack) == 22
    assert len(stack.remove_list) == 0
    assert len(stack.connected_substacks) == 2 

def test_get_scene_ids(stack):
    stack.remove_pairs(stack.full_stack[1:11])
    largest_fully_connected_subset_stack_ids = stack.get_scene_ids()
    remove_list_scene_ids = stack.get_scene_ids(stack.remove_list)
    full_stack_scene_ids = stack.get_scene_ids(stack.full_stack)
    subset_stack_scene_ids = stack.get_scene_ids(stack.subset_stack)
    assert len(largest_fully_connected_subset_stack_ids) == 10
    assert len(remove_list_scene_ids) == 10
    assert len(full_stack_scene_ids) == 21
    assert len(subset_stack_scene_ids) == 11

def test_stack_disallow_missing_state_vectors(stack_results):
    """
    Test the optional allow_missing_state_vectors argument set to False (default). 
    """
    stack_results[2].baseline["stateVectors"]["positions"]["prePositionTime"] = None
    missing_state_vectors_not_allowed_stack = Stack.from_search_results(stack_results)
    assert len(missing_state_vectors_not_allowed_stack.full_stack) == 15
    assert len([p for p in missing_state_vectors_not_allowed_stack.full_stack if p.perpendicular_baseline is None]) == 0

def test_stack_allow_missing_state_vectors(stack_results):
    """
    Test the optional allow_missing_state_vectors argument set to False (default). 
    """
    stack_results[2].baseline["stateVectors"]["positions"]["prePositionTime"] = None
    missing_state_vectors_allowed_stack = Stack.from_search_results(stack_results, allow_missing_state_vectors=True)
    assert len(missing_state_vectors_allowed_stack.full_stack) == 21
    assert len([p for p in missing_state_vectors_allowed_stack.full_stack if p.perpendicular_baseline is None]) == 6



from asf_search.CMR.MissionList import get_campaigns
from asf_search.search.campaigns import _get_project_names
import pytest
import requests_mock

from asf_search.constants.INTERNAL import CMR_COLLECTIONS_PATH, CMR_HOST
from asf_search.exceptions import CMRError


def test_getMissions_error():
    with requests_mock.Mocker() as m:
        m.register_uri(
            'POST',
            'https://' + CMR_HOST + CMR_COLLECTIONS_PATH,
            status_code=300,
            json={'error': {'report': ''}},
        )

        with pytest.raises(CMRError):
            get_campaigns({})


def test_getMissions_error_parsing():
    with requests_mock.Mocker() as m:
        m.post('https://' + CMR_HOST + CMR_COLLECTIONS_PATH)

        with pytest.raises(CMRError):
            get_campaigns({})


def run_test_get_project_names(cmr_ummjson, campaigns):
    assert _get_project_names(cmr_ummjson) == campaigns



from asf_search import ASF_LOGGER, ASFSearchOptions
from asf_search import INTERNAL
import requests



def report_search_error(search_options: ASFSearchOptions, message: str):
    """Reports CMR Errors automatically to ASF"""

    from asf_search import REPORT_ERRORS

    if not REPORT_ERRORS:
        ASF_LOGGER.warning(
            'Automatic search error reporting is turned off,'
            'search errors will NOT be reported to ASF.'
            '\nTo enable automatic error reporting, set asf_search.REPORT_ERRORS to True'
            '\nIf you have any questions email uso@asf.alaska.edu'
        )
        return

    user_agent = search_options.session.headers.get('User-Agent')
    search_options_list = '\n'.join(
        [f'\t{option}: {key}' for option, key in dict(search_options).items()]
    )
    message = f'Error Message: {str(message)}\nUser Agent: {user_agent} \
    \nSearch Options: {{\n{search_options_list}\n}}'

    response = requests.post(
        f'https://{INTERNAL.ERROR_REPORTING_ENDPOINT}',
        data={'Message': f'This error message and info was automatically generated:\n\n{message}'},
    )

    try:
        response.raise_for_status()
    except requests.exceptions.HTTPError:
        ASF_LOGGER.error(
            'asf-search failed to automatically report an error,'
            'if you have any questions email uso@asf.alaska.edu'
            f"\nError Text: HTTP {response.status_code}: {response.json()['errors']}"
        )
        return
    if response.status_code == 200:
        ASF_LOGGER.error(
            (
                'The asf-search module ecountered an error with CMR,'
                'and the following message was automatically reported to ASF:'
                '\n\n"\nmessage\n"'
                'If you have any questions email uso@asf.alaska.edu'
            )
        )



from collections import defaultdict, deque
from copy import copy
from typing import Optional, List, Tuple, Union
import warnings

from .ASFProduct import ASFProduct
from .Pair import Pair
from .ASFSearchOptions import ASFSearchOptions
from .ASFSearchResults import ASFSearchResults
from .warnings import PairNotInFullStackWarning

class Stack:
    """
    A Stack object contains 4+ lists of Pair objects. Each Pair contains a pair of asf_search.ASFProduct objects.
    
    Stack member variables holding lists of Pairs:
    - Stack.full_stack: Every possible pair based on the provided geo_reference scene 
      and ASFSearchOptions. This forms a complete network of all represented asf_search.ASFProducts.
    - Stack.remove_list: The list of Pairs to remove from Stack.full_stack, used to create Stack.subset_stack
    - Stack.subset_stack: The resulting list after removing Stack.Remove_list from Stack.full_stack. This creates 
      a possibly disconnected network of asf_search.ASFProducts
    - Stack.connected_substacks: This is a list of lists of Pairs. It contains each disconnected component of a Stack.subset_stack's
      asf_search.ASFProduct network. A length of 1 indicates that Stack.subset_stack is a connected network of asf_search.ASFProducts.

    Public Stack methods:
    - Stack.remove_pairs(): Adds Pairs to Stack.remove_list and removes them from Stack.subset_stack.
    - Stack.add_pairs(): Adds Pairs to Stack.subset_stack. This either removes them from Stack.remove_list, or if not yet present in the
      Stack, adds them to self.full_stack.
    - Stack.get_scene_ids(): A convenience method that returns a given list of Pairs as a list of tuples of asf_search.ASFProduct
      product IDs, which is useful when ordering on-demand processing via ASF's HyP3 or HyP3+ services.
    """
    def __init__(
        self,
        geo_reference: ASFProduct,
        opts: Optional[ASFSearchOptions] = None,
        allow_missing_state_vectors: Optional[bool] = False
    ):
        """
        Constructor that builds a Stack from a geo-reference ASFProduct

        geo_reference: An ASFProduct that serves as a geo-reference scene for the Stack
        opts: (Optional) ASFSearchOptions to apply to the geo_reference.stack() search when creating Stack.full_stack
        """
        self.geo_reference = geo_reference
        if opts is None:
            opts = ASFSearchOptions()
        self.opts = opts
        self.allow_missing_state_vectors = allow_missing_state_vectors
        self.full_stack = self._build_full_stack()
        self._remove_list = []
        self._update_stack()

    @classmethod
    def from_search_results(
        cls,
        stack_search_results: ASFSearchResults,
        allow_missing_state_vectors: Optional[bool] = False
    ):
        """
        Alternate class method constructor using ASFSearchResults instead of a single geo_reference.
        """
        obj = cls.__new__(cls)

        obj.allow_missing_state_vectors = allow_missing_state_vectors
        obj.full_stack = obj._build_full_stack(stack_search_results)
        obj._remove_list = []
        obj.subset_stack = obj._get_subset_stack()
        obj.connected_substacks = obj._find_connected_substacks()
        obj.geo_reference = None

        return obj

    @property
    def remove_list(self) -> List[Pair]:
        """
        Returns a copy of self._remove_list so client changes 
        do not alter self._remove_list without initiating a stack update

        Disallow: 
          - my_stack.remove_list.append(my_pair)
          - my_stack.remove_list.remove(my_pair)

        Support:
          - my_stack.remove_pairs([pair_1, pair_2, ...])
          - my_stack.add_pairs([pair_1, pair_2, ...])
        """
        return copy(self._remove_list)

    @remove_list.setter
    def remove_list(self, pairs: List[Pair]):
        """
        pairs: A list of Pairs to remove from self.subset_stack
        """
        # remove duplicates
        self._remove_list = list(set(pairs))
        self._update_stack()

    def remove_pairs(self, pairs: Union[List[Pair], Pair]):
        """
        Remove pairs from self.subset_stack, 
        i.e., add them to self._remove_list

        pairs: A Pair or list of Pairs to remove from self.subset_stack
        """
        if isinstance(pairs, Pair):
            pairs = [pairs]

        for pair in pairs:
            if pair not in self._remove_list:
                if pair in self.full_stack:
                    self._remove_list.append(pair)
                else:
                    msg = f"warning: {pair} is not in full_stack"
                    warnings.warn(PairNotInFullStackWarning(msg))
        self._update_stack()

    def add_pairs(self, pairs: Union[List[Pair], Pair]):
        """
        Add pairs to self.subset_stack and, if necessary, to self.full_stack 
        i.e., remove them from self._remove_list if present or else add them to self.full_stack 

        This allows for the addition of custom pairs that were not originally present
        in self.full_stack

        pairs: A Pair or list of Pairs to add to self.subset_stack
        """
        if isinstance(pairs, Pair):
            pairs = [pairs]

        for pair in pairs:
            if pair in self._remove_list:
                self._remove_list.remove(pair)
            else:
                self.full_stack.append(pair)
        self._update_stack()

    def _build_full_stack(self, stack_search_results: Optional[ASFSearchResults]=None) -> List[Pair]:
        """
        Create self._full_stack, which involves performing a stack search
        of the georeference scene and creating a list of every possible Pair.

        stack_search_results: (Optional) ASFSearchResults from an ASFProduct.stack search
        """
        if stack_search_results is None: 
            stack_search_results = self.geo_reference.stack(opts=self.opts)

        return [
            Pair(p1, p2)
            for i, p1 in enumerate(stack_search_results)
            for p2 in stack_search_results[i + 1:]
            if self.allow_missing_state_vectors or Pair(p1, p2).perpendicular_baseline is not None
        ]

    def _get_subset_stack(self) -> List[Pair]:
        """
        Create a subset_stack by removing every pair in
        self.remove_list from self.full_stack
        """
        return [pair for pair in self.full_stack if pair not in self.remove_list]

    def _update_stack(self):
        """
        Recalculate self.subset_stack and find its connected substacks.
        These two things should always happen together.
        """
        self.subset_stack = self._get_subset_stack()
        self.connected_substacks = self._find_connected_substacks()

    def _find_connected_substacks(self) -> List[List[Pair]]:
        """
        Perform a bredth first search to find all connected components of self.subset_stack
        """

        graph = defaultdict(list)
        for pair in self.subset_stack:
            graph[pair.ref].append(pair.sec)
            graph[pair.sec].append(pair.ref)

        visited_nodes = set()
        visited_pairs = set()
        components = []

        for node in graph:
            if node not in visited_nodes:
                component_nodes = set()
                component_pairs = {}

                queue = deque([node])
                visited_nodes.add(node)

                while queue:
                    current = queue.popleft()
                    component_nodes.add(current)

                    for neighbor in graph[current]:
                        if (current, neighbor) not in visited_pairs and (neighbor, current) not in visited_pairs:
                            for pair in self.subset_stack:
                                if (pair.ref == current and pair.sec == neighbor) or \
                                    (pair.sec == current and pair.ref == neighbor):
                                    component_pairs[Pair(pair.ref, pair.sec)] = pair
                                    break
                            visited_pairs.add((current, neighbor))
                            visited_pairs.add((neighbor, current))

                        if neighbor not in visited_nodes:
                            visited_nodes.add(neighbor)
                            queue.append(neighbor)
                component_pairs = [v for v in component_pairs.values()]
                components.append(component_pairs)

        return components

    def get_scene_ids(self, pair_list: Optional[List[Pair]] = None) -> List[Tuple[str, str]]:
        """
        Provides scene names for all asf_search.ASFProducts in a list of Pairs.
        Useful when ordering pair-based products from ASF HyP3 On-Demand Processing.

        If no stack_dict is passed, defaults to the largest connected substack

        pair_list: (Optional) A list of `Pair`s for which to retrieve scene IDs.
        
        Returns:
            A list tuples containing the reference and secondary scene names for each `Pair` in a `Pair` list
        """
        if not pair_list:
            pair_list = max(self.connected_substacks, key=len)

        return [
            (pair.ref.properties["sceneName"], pair.sec.properties["sceneName"])
            for pair in pair_list
            ]
    


from datetime import datetime, timedelta
import importlib.util
import numpy as np
from typing import Optional, List, Dict, Tuple, Union

from .ASFProduct import ASFProduct
from .Pair import Pair
from .Stack import Stack
from .ASFSearchOptions import ASFSearchOptions
from .ASFSearchResults import ASFSearchResults

try:
    from ciso8601 import parse_datetime
except ImportError:
    from dateutil.parser import parse as parse_datetime

_SBASNETWORK_PLOT_OPT_DEPS = ['plotly', 'networkx', 'pandas']
try:
    for spec in _SBASNETWORK_PLOT_OPT_DEPS:
        if importlib.util.find_spec(spec) is None:
            raise ImportError
        
    import plotly.graph_objects as go
    import networkx as nx
    from plotly.colors import sample_colorscale
    import pandas as pd
except ImportError:
    go = None
    nx = None
    sample_colorscale = None
    pd = None

class SBASNetwork(Stack):
    """
    SBASNetwork is a child class of Stack, used to create seasonal SBAS networks. It can be used
    to create multiannual networks connected with multiannual bridge pairs.
    """
    def __init__(
        self,
        geo_reference: ASFProduct = None,
        start_date: str = None,
        end_date: str = None,
        season: Tuple[int] = (1, 365),
        perpendicular_baseline: Optional[int] = 400,
        inseason_temporal_baseline: Optional[int] = 36,
        bridge_year_threshold: Optional[int] = 1,
        bridge_target_date: Optional[str] = None,
        opts: Optional[ASFSearchOptions] = ASFSearchOptions(),
        allow_missing_state_vectors: Optional[bool] = False
    ):
        """
        geo_reference: an ASFProduct used as a georeference scene for the network
        perpendicular_baseline: the perpendicular baseline for the SBAS network
        inseason_temporal_baseline: the temporal baseline for the SBAS network, not accounting for multiannual bridging
        bridge_year_threshold: the number of year for which to allow multiannual bridging
        bridge_target_date: %m-%d string of the inseason bridge date to target. 
                            Reference scenes for bridge pairs are valid within an inseason_temporal_baseline of this date.
        opts: ASFSearchOptions to limit the size of your SBASNetwork and define seasons
              Recommended to include: "start", "end", "season"
        """
        self.season = season
        if bridge_target_date:
            self.bridge_target_date = bridge_target_date
        else:
            self.bridge_target_date = self.season[0]
        self.start_date = start_date
        self.end_date = end_date
        self.perpendicular_baseline = perpendicular_baseline
        self.inseason_temporal_baseline = inseason_temporal_baseline
        self.bridge_year_threshold = bridge_year_threshold
        opts.merge_args(
            start=self.start_date,
            end=self.end_date,
            season=self.season,
        )
        self.opts = opts
        self.allow_missing_state_vectors = allow_missing_state_vectors

        self.temporal_baseline = (bridge_year_threshold * 365) + inseason_temporal_baseline

        super().__init__(
            geo_reference=geo_reference, 
            opts=self.opts,
            allow_missing_state_vectors = self.allow_missing_state_vectors
            )

        self._build_sbas_network()

    @classmethod
    def from_search_results(
        cls,
        stack_search_results: ASFSearchResults,
        season: Tuple[int] = (1, 365),
        perpendicular_baseline: Optional[int] = 400,
        inseason_temporal_baseline: Optional[int] = 36,
        bridge_year_threshold: Optional[int] = 1,
        bridge_target_date: Optional[str] = None,
        allow_missing_state_vectors: Optional[bool] = False
    ):
        """
        Alternate class method constructor using ASFSearchResults instead of a single geo_reference.
        """
        obj = cls.__new__(cls)

        obj.season = season
        if bridge_target_date:
            obj.bridge_target_date = bridge_target_date
        else:
            obj.bridge_target_date = obj.season[0]
        dates = [parse_datetime(product.properties["startTime"]).date() for product in stack_search_results]
        obj.start_date = str(min(dates))
        obj.end_date = str(max(dates))
        obj.perpendicular_baseline = perpendicular_baseline
        obj.inseason_temporal_baseline = inseason_temporal_baseline
        obj.bridge_year_threshold = bridge_year_threshold
        obj.allow_missing_state_vectors = allow_missing_state_vectors
        obj.full_stack = obj._build_full_stack(stack_search_results)
        obj._remove_list = []
        obj.subset_stack = obj._get_subset_stack()
        obj.connected_substacks = obj._find_connected_substacks()
        obj.geo_reference = None

        obj._build_sbas_network()

        return obj

    def _build_full_stack(self, stack_search_results: Optional[ASFSearchResults]=None) -> List[Pair]:
        """
        Create self._full_stack, which involves performing a stack search
        of the georeference scene and creating a list of every possible Pair.

        Overrides Stack._build_full_stack() to naively exclude pairs based on self.perpendicular_baseline,
        self.inseason_temporal_baseline, and self.bridge_year_threshold. This is done for efficiency, to
        avoid build massive full_stacks. More targeted occures in self._build_sbas_network()

        stack_search_results: (Optional) ASFSearchResults from an ASFProduct.stack search

        Returns: A list of pairs filtered by baselines, and including every possible bridge pair within 
                 self.bridge_year_threshold
        """
        if stack_search_results is None: 
            stack_search_results = self.geo_reference.stack(opts=self.opts)

        if self.allow_missing_state_vectors:
             full_stack = [
                Pair(p1, p2)
                for i, p1 in enumerate(stack_search_results)
                for p2 in stack_search_results[i+1:]
                if (Pair(p1, p2).perpendicular_baseline is None
                or Pair(p1, p2).perpendicular_baseline <= self.perpendicular_baseline)
                and Pair(p1, p2).temporal_baseline.days <= self.inseason_temporal_baseline + (self.bridge_year_threshold * 365)
            ]
        else:
            full_stack = [
                Pair(p1, p2)
                for i, p1 in enumerate(stack_search_results)
                for p2 in stack_search_results[i+1:]
                if Pair(p1, p2).perpendicular_baseline is not None
                and Pair(p1, p2).perpendicular_baseline <= self.perpendicular_baseline
                and Pair(p1, p2).temporal_baseline.days <= self.inseason_temporal_baseline + (self.bridge_year_threshold * 365)
            ]
        return full_stack

    def _is_valid_bridge_pair(self, pair: Pair) -> bool:
        """
        Logic to determine if a pair meets the criteria for including in an SBAS network as a bridge pair:
        - reference scene date within 1/2 self.inseason_temporal_baseline on either side of self.bridge_target_date
        - secondary scene within 1/2 self.inseason_temporal_baseline on either side of self.bridge_target_date for 
          each year within the self.bridge_year_threshold

        pair: the Pair to check
        """
        # return True right away if pair within inseason temp baseline
        if pair.temporal_baseline.days <= self.inseason_temporal_baseline:
            return True

        # determine how far ref scene date is from target multi-annual bridge date 
        days_from_bridge_date = np.abs(
            (
                datetime.strptime(f"{self.bridge_target_date}-{pair.ref_time.year}", "%m-%d-%Y") 
                - pair.ref_time.replace(tzinfo=None)).days
            )
        
        half_inseason_temporal_baseline = self.inseason_temporal_baseline / 2

        # Create lists of valid secondary scene date ranges.
        # The number of years bridged is determined by the bridge_year_threshold.
        valid_ranges = []
        for i in range(1, self.bridge_year_threshold+1):
            valid_ranges.append((i * 365 - half_inseason_temporal_baseline, i * 365 + half_inseason_temporal_baseline))

        # return True if the ref scene is within the inseason_temporal_baseline / 2 on either
        # side of the target bridge date and the secondary scene falls within a valid date range
        return days_from_bridge_date <= half_inseason_temporal_baseline and \
            any(start <= pair.temporal_baseline.days <= end for start, end in valid_ranges)

    def _build_sbas_network(self):
        """
        Create a self._remove_list for filtering self.full_stack to produce self.subset_stack.
        Initial temporal and perpendicular baselng occur in SBASNetwork._build_full_stack().
        This method appies additional filtering to remove potential pairs based on self.inseason_temporal_baseline
        and self.bridge_year_threshold
        """
        remove_list = []
        for pair in self.full_stack:
            if not self._is_valid_bridge_pair(pair):
                remove_list.append(pair)
        self.remove_list = remove_list

    def add_digraph_edge_traces(self, digraph, largest_network, pair_lists, node_positions):
        """
        Used for SBASNetwork plotting. Adds networkx traces to graph edges

        digraph: networkx.DiGraph
        largest_network: The largest (perhaps only) connected SBASNetwork pair list being plotted
        pair_lists: list of all pair lists being plotted
        node_positions: Tuple[Dict[networkx.DiGraph.node, datetime.timestamp], Dict[networkx.DiGraph.node, float]]
                        representing the x and y positions for each node    

        returns: A plotly.graph_objects.Scatter object for graph edges   
        """
    
        insar_node_pairs = [
            (
                datetime.strftime(pair.ref_time, "%Y-%m-%d"),
                datetime.strftime(pair.sec_time, "%Y-%m-%d"),
                {
                    "perp_bs": pair.perpendicular_baseline,
                    "temp_bs": pair.temporal_baseline.days
                }
            )
            for pair in largest_network
        ]
        digraph.add_edges_from(insar_node_pairs, data=True)

        edge_x = []
        edge_y = []
        edge_text = []
        for edge in digraph.edges(data=True):
            x0 = node_positions[0][edge[0]]
            y0 = node_positions[1][edge[0]]
            x1 = node_positions[0][edge[1]]
            y1 = node_positions[1][edge[1]]
            edge_x.extend([x0, x1, None])
            edge_y.extend([y0, y1, None])
            edge_text.append(
                f"{edge[0]} - {edge[1]}, perp baseline: {edge[2]['perp_bs']}, temp baseline: {edge[2]['temp_bs']}"
            )

        edge_traces = [
            go.Scatter(
                x=edge_x,
                y=edge_y,
                line=dict(width=4, color="rgba(52, 114, 168, 0.7)"),  # blue
                mode="lines",
            ),
            go.Scatter(
                x=[
                    (node_positions[0][edge[0]] + node_positions[0][edge[1]]) / 2
                    for edge in digraph.edges()
                ],
                y=[
                    (node_positions[1][edge[0]] + node_positions[1][edge[1]]) / 2
                    for edge in digraph.edges()
                ],
                mode="markers",
                marker=dict(size=20, color="rgba(255, 255, 255, 0)"),
                hoverinfo="text",
                text=edge_text,
            )
        ]

        # Add additional sub-networks
        other_networks = [s for s in pair_lists if s is not largest_network]
        colors = get_n_colors(len(other_networks))
        for color, other_network in zip(colors, other_networks):
            edge_x = []
            edge_y = []
            for pair in other_network:
                ref = datetime.strftime(pair.ref_time, "%Y-%m-%d")
                sec = datetime.strftime(pair.sec_time, "%Y-%m-%d")
                if ref in node_positions[0] and sec in node_positions[0]:
                    edge_x.extend([node_positions[0][ref], node_positions[0][sec], None])
                    edge_y.extend([node_positions[1][ref], node_positions[1][sec], None])
            edge_traces.append(go.Scatter(
                x=edge_x,
                y=edge_y,
                line=dict(width=4, color=color),
                mode="lines",
            ))

            edge_text = []
            hover_x = []
            hover_y = []
            for pair in other_network:
                ref = datetime.strftime(pair.ref_time, "%Y-%m-%d")
                sec = datetime.strftime(pair.sec_time, "%Y-%m-%d")
                if ref in node_positions[0] and sec in node_positions[0]:
                    edge_text.append(
                        f"{ref} - {sec}, perp baseline: {pair.perpendicular_baseline}, temp baseline: {pair.temporal_baseline.days}"
                    )
                    hover_x.append((node_positions[0][ref] + node_positions[0][sec]) / 2)
                    hover_y.append((node_positions[1][ref] + node_positions[1][sec]) / 2)

            edge_traces.append(go.Scatter(
                x=hover_x,
                y=hover_y,
                mode="markers",
                marker=dict(size=20, color="rgba(255, 255, 255, 0)"),
                hoverinfo="text",
                text=edge_text,
            ))
        return edge_traces

    def add_digraph_node_traces(self, digraph, plot_geo_ref, pair_lists, node_positions):
        """
        Used for SBASNetwork plotting. Adds networkx traces to nodes of used and unused SLCs

        digraph: networkx.DiGraph
        plot_geo_ref: The ASFProduct being used as the geographic reference scene for the plot
        pair_lists: list of all pair lists being plotted
        node_positions: Tuple[Dict[networkx.DiGraph.node, datetime.timestamp], Dict[networkx.DiGraph.node, float]]
                        representing the x and y positions for each node

        returns: Tuple of plotly.graph_objects.Scatter objects for used and unused SLCs
        """
        # Add used scene node locations
        node_x = [node_positions[0][node] for node in digraph.nodes()]
        node_y = [node_positions[1][node] for node in digraph.nodes()]
        node_text = [digraph.nodes[node]["date"] for node in digraph.nodes()]

        node_traces = go.Scatter(
            x=node_x,
            y=node_y,
            mode="markers+text",
            textposition="top center",
            marker=dict(size=15, color="rgba(32, 33, 32, 0.7)", line_width=0),
            hoverinfo="text",
            hovertext=node_text,
        )

        # Add unused scene node locations
        all_slcs = set(scene for pair in self.full_stack for scene in (pair.ref, pair.sec))
        used_slcs = set()
        for sd in pair_lists:
            used_slcs.update(scene for pair in sd for scene in (pair.ref, pair.sec))
        unused_slcs = list(all_slcs - used_slcs)

        unused_slc_dates_str = [i.properties["stopTime"].split("T")[0] for i in unused_slcs]
        unused_slc_dates_x = [datetime.strptime(i, "%Y-%m-%d").timestamp() for i in unused_slc_dates_str]
        unused_slc_perp_y = [Pair(plot_geo_ref, i).perpendicular_baseline for i in unused_slcs]

        unused_slcs_traces = go.Scatter(
            x=unused_slc_dates_x, y=unused_slc_perp_y,
            mode='markers',
            hoverinfo='text',
            text=unused_slc_dates_str,
            marker=dict(
                color='rgba(255,0,0,0.5)',
                size=10,
                line_width=2
            )
        )
        return node_traces, unused_slcs_traces
    
    def build_plot_header_text(self, plot_geo_ref: ASFProduct, largest_network: List[Pair], pair_list: Union[List[Pair], None]) -> str:
        """
        Used for SBASNetwork plotting. Creates the header text for a SBASNetwork plot

        plot_geo_ref: The ASFProduct being used as the geographic reference scene for the plot
        largest_network: The largest (perhaps only) connected SBASNetwork pair list being plotted
        pair_list: The value of the optional pair_list arg passed to plot (could be None)

        Returns: Header text string
        """
        largest_network_slc_count = len(set(scene for pair in largest_network for scene in (pair.ref, pair.sec)))

        if pair_list is self.full_stack:
            plot_header_text = (
                "<b>SBAS Stack</b><br>"
                f"Geographic Reference: {plot_geo_ref.properties['sceneName']}<br>"
                f"Temporal Bounds: {self.start_date.split('T')[0]} - {self.end_date.split('T')[0]}, "
                f"Seasonal Bounds: {julian_to_month_day(plot_geo_ref, self.season)}<br>"
                f"Full Stack Size: {len(largest_network)} pairs from {largest_network_slc_count} scenes<br>"
            )
        else:
            plot_header_text = (
                "<b>SBAS Stack</b><br>"
                f"Geographic Reference: {plot_geo_ref.properties['sceneName']}<br>"
                f"Temporal Bounds: {self.start_date.split('T')[0]} - {self.end_date.split('T')[0]}, "
                f"Seasonal Bounds: {julian_to_month_day(plot_geo_ref, self.season)}<br>"
                f"Temporal Baseline: {self.inseason_temporal_baseline} days, "
                f"Perpendicular Baseline: {self.perpendicular_baseline}m<br>"
                f"Bridge Target Date: {self.bridge_target_date}, "
                f"Bridge Year Threshold: {self.bridge_year_threshold}, Largest Stack Size: "
                f"{len(largest_network)} pairs from {largest_network_slc_count} scenes<br>"
            )
        return plot_header_text

    def plot(self, pair_list: Optional[Union[List[Pair], None]] = None):
        """
        Plot the SBAS network(s). Accepts a pair_list or a list of pair_lists.
        The largest network is plotted in blue; others are plotted in distinct colors.
        Possible member networks to pass as pair_list are: 
            - `self.full_stack`: an SBAS network filtered by baselines, and including every possible bridge pair within 
               self.bridge_year_threshold
            - `self.subset_stack`: a possibly disconnected SBAS network filtered by baselines, with additional 
               bridge pair filtering applied
            - `self.connected_substacks`: self.subset_stack, broken up into a list of connected sub-networks

        pair_list (optional): One of the SBASNetwork's pair lists or lists of pair list (options listed above).
                               Default: `self.connected_substacks`.
        """

        if go is None or nx is None or sample_colorscale is None or pd is None:
            raise ImportError(
                'The `plot()` method requires the optional asf-search '
                f'dependencies {_SBASNETWORK_PLOT_OPT_DEPS}, '
                'but they could not be found in the current python environment. '
                'Enable this method by including the appropriate pip or conda install. '
                'Ex: `python -m pip install asf-search[sbasnetwork_plot]`'
            )

        pair_lists = self.connected_substacks if not pair_list else [pair_list]
        largest_network = max(pair_lists, key=lambda s: len(s))
        node_products = build_node_products(pair_lists)
        G = nx.DiGraph()
        plot_geo_ref = self.geo_reference if self.geo_reference is not None else self.subset_stack[0].ref
        add_digraph_nodes(G, plot_geo_ref, node_products)
        node_x_positions, node_y_positions = get_node_positions(G, plot_geo_ref, node_products)
        edge_traces = self.add_digraph_edge_traces(G, largest_network, pair_lists, (node_x_positions, node_y_positions))
        node_traces, unused_slcs_traces = self.add_digraph_node_traces(G, plot_geo_ref, pair_lists, (node_x_positions, node_y_positions))

        start_date, end_date = get_pair_lists_date_range(pair_lists)
        date_range = (
            pd.date_range(
                start=start_date,
                end=end_date,
                freq="MS"
            )
            .strftime("%Y-%m")
            .tolist()
        )
        date_range_ts = [datetime.strptime(date, "%Y-%m").timestamp() for date in date_range]

        fig = go.Figure(
            data=edge_traces + [node_traces, unused_slcs_traces],
            layout=go.Layout(
                showlegend=False,
                hovermode="closest",
                height=800,
                margin=dict(t=180),
                xaxis=dict(
                    title="Acquisition Date",
                    tickvals=date_range_ts,
                    ticktext=date_range,
                    gridcolor="gray",
                    zerolinecolor="gray",
                ),
                yaxis=dict(
                    title="Perpendicular Baseline (m)",
                    gridcolor="gray",
                    zerolinecolor="gray",
                ),
                title=dict(
                    text=self.build_plot_header_text(plot_geo_ref, largest_network, pair_list),
                    y=0.95,
                    x=0.5,
                    xanchor="center",
                    yanchor="top",
                    font=dict(family="Helvetica, monospace", size=22),
                ),
                plot_bgcolor="white",
                paper_bgcolor="lightgrey",
            ),
        )
        fig.show()

    def get_pair_from_dates(self, pair_list: List[Pair],
                                ref_date: datetime.date, sec_date: datetime.date) -> Pair:
        """
        This convenience method allows for the retrieval of a Pair object from a list
        of Pairs by reference and secondary date. This is usefull when identifying a 
        Pair object for removal or addition to an SBASNetwork.subset_stack

        This method is included in SBASNetwork and not Stack because we only care about
        dates in SBASNetwork and Stack is aware of datetimes.

        pair_list: a list of Pairs from which to find a Pair object
        ref_date: a datetime.date of a target Pair's reference scene
        sec_date: a datetime.date of a target Pair's secondary scene

        Returns: A Pair with a corresponding ref_date and sec_date
        """
        for pair in pair_list:
            if pair.ref_time.date() == ref_date and pair.sec_time.date() == sec_date:
                return pair

def get_n_colors(n: int, colorscale: str="Rainbow", alpha: float=0.4) -> List[str]:
    """
    Used for SBASNetwork plotting. Creates an n-length list of rgba colors from a plotly colorscale

    n: The length of the list of colors to return
    colorscale: the plotly sample_colorscale to from which to extract rgba colors
    alpha: The opacity to apply to the returned rgba colors

    returns: An n-length list of rgba colors from the sample_coloscale
    """
    base_colors = sample_colorscale(colorscale, [i / max(n - 1, 1) for i in range(n)])
    rgba_colors = []
    for color_str in base_colors:
        if color_str.startswith("rgb"):
            nums = color_str.strip("rgb()").split(",")
            r, g, b = [int(x) for x in nums]
            rgba_colors.append(f"rgba({r}, {g}, {b}, {alpha})")
        else:
            raise ValueError(f"Unexpected color format: {color_str}")
    return rgba_colors

def build_node_products(pair_lists: List[List[Pair]]) -> Dict[str, ASFProduct]:
    """
    Used for SBASNetwrok plotting. Builds a dictionary of all ASFProducts in
    a list of lists of Pairs

    pair_lists: A list of lists of Pairs

    returns: A dictionary with date strings as keys and ASFProducts as values for
             every product in the passed Pair lists
    """
    node_products = {}
    for network in pair_lists:
        for pair in network:
            for date_obj, product in [(pair.ref_time, pair.ref), (pair.sec_time, pair.sec)]:
                date_str = datetime.strftime(date_obj, "%Y-%m-%d")
                if date_str not in node_products:
                    node_products[date_str] = product
    return node_products

def add_digraph_nodes(digraph, plot_geo_ref, node_products):
    """
    Used for SBASNetwork plotting. Adds nodes to digraph.

    digraph: networkx.DiGraph
    plot_geo_ref: ASFProduct being used as a geographic reference for the network plot
    node_products: Dict[str, ASFProduct] containing date strings as keys and ASFProducts as values

    """
    for date_str, product in node_products.items():
        digraph.add_node(date_str)
        digraph.nodes[date_str]["date"] = date_str
        digraph.nodes[date_str]["perp_bs"] = Pair(plot_geo_ref, product).perpendicular_baseline

def get_node_positions(digraph, plot_geo_ref, node_products):
    """
    Used for SBASNetwork plotting. Determines x/y node coordinates for digraph.

    digraph: networkx.DiGraph
    plot_geo_ref: ASFProduct being used as a geographic reference for the network plot
    node_products: Dict[str, ASFProduct] containing date strings as keys and ASFProducts as values

    returns: Tuple[Dict[networkx.DiGraph.node, datetime.timestamp], Dict[networkx.DiGraph.node, float]].
             node_x_positions is a dictionary with nodes as values and and ASFProducts' datetime.timestamps as values.
             node_y_positions is a dictionary with nodes as values and and ASFProducts's perpendicular baselines 
             relative to plot_geo_ref as values.
    """
    for date_str, product in node_products.items():
        digraph.add_node(date_str)
        digraph.nodes[date_str]["date"] = date_str
        digraph.nodes[date_str]["perp_bs"] = Pair(plot_geo_ref, product).perpendicular_baseline
    node_x_positions = {
        node: datetime.strptime(data["date"], "%Y-%m-%d").timestamp()
        for node, data in digraph.nodes(data=True)
    }
    node_y_positions = {
        node: (data["perp_bs"] if data.get("perp_bs") else 0)
        for node, data in digraph.nodes(data=True)
    }
    return node_x_positions, node_y_positions

def get_pair_lists_date_range(pair_lists: List[List[Pair]]) -> Tuple[str, str]:
    """
    Used for SBASNetwork plotting. Finds the acquisition date range from a list
    of lists of Pairs

    pair_lists: A list of lists of Pairs

    returns: The acquisition date range covered by the ASFProducts in pair_lists' Pairs
    """
    start_dates, end_dates = [], []
    for pair_list in pair_lists:
        start_dates.append(min([pair.ref_time for pair in pair_list]))
        end_dates.append(max([pair.sec_time for pair in pair_list]))

    start_date = min(start_dates)
    start_date_str = f"{start_date.year}-{start_date.month}"
    end_date = max(end_dates)
    end_date_str = f"{end_date.year}-{end_date.month}"

    return start_date_str, end_date_str

def julian_to_month_day(ref_product: ASFProduct, julian_tuple: Tuple[int, int]) -> Tuple[str, str]:
    """
    Used for SBASNetwork plotting. Converts a tuple of julian days to month-day strings

    ref_product: an ASFProduct from which to extract a year when converting julian days
    julian_tuple: a tuple of two julian days

    returns: A tuple of month-day strings cooresponding to the passed julian days in the 
             year of ref_product's acquisition
    """
    year = int(ref_product.properties['stopTime'].split('-')[0])
    month_day = []
    for day in julian_tuple:
        date = datetime(year, 1, 1) + timedelta(days=day - 1)
        month_day.append(date.strftime("%m-%d"))
    return tuple(month_day)


from typing import Dict, Union
from asf_search import ASFSession, ASFStackableProduct
from asf_search.CMR.translate import try_parse_float, try_parse_int
from asf_search.constants import PRODUCT_TYPE


class RADARSATProduct(ASFStackableProduct):
    """
    ASF Dataset Documentation Page: https://asf.alaska.edu/datasets/daac/radarsat-1/
    """

    _base_properties = {
        **ASFStackableProduct._base_properties,
        'faradayRotation': {'path': ['AdditionalAttributes', ('Name', 'FARADAY_ROTATION'), 'Values', 0], 'cast': try_parse_float},
        'md5sum': {'path': ['AdditionalAttributes', ('Name', 'MD5SUM'), 'Values', 0]},
        'beamModeType': {'path': ['AdditionalAttributes', ('Name', 'BEAM_MODE_TYPE'), 'Values', 0]},
        'insarStackId': {'path': ['AdditionalAttributes', ('Name', 'INSAR_STACK_ID'), 'Values', 0]},
        'frameNumber': {'path': ['AdditionalAttributes', ('Name', 'FRAME_NUMBER'), 'Values', 0], 'cast': try_parse_int}, #Sentinel and ALOS product alt for frameNumber (ESA_FRAME)
        'esaFrame': {'path': ['AdditionalAttributes', ('Name', 'CENTER_ESA_FRAME'), 'Values', 0], 'cast': try_parse_int},
    }

    def __init__(self, args: Dict = {}, session: ASFSession = ASFSession()):
        super().__init__(args, session)

    @staticmethod
    def get_default_baseline_product_type() -> Union[str, None]:
        """
        Returns the product type to search for when building a baseline stack.
        """
        return PRODUCT_TYPE.L0



from typing import Dict
from asf_search import ASFSession, ASFProduct
from asf_search.CMR.translate import try_parse_int, try_round_float


class SEASATProduct(ASFProduct):
    """
    ASF Dataset Documentation Page: https://asf.alaska.edu/data-sets/sar-data-sets/seasat/
    """

    _base_properties = {
        **ASFProduct._base_properties,
        'md5sum': {'path': ['AdditionalAttributes', ('Name', 'MD5SUM'), 'Values', 0]},
        'frameNumber': {'path': ['AdditionalAttributes', ('Name', 'FRAME_NUMBER'), 'Values', 0], 'cast': try_parse_int}, # for consolidated collection
        'bytes': {'path': ['DataGranule', 'ArchiveAndDistributionInformation']},
    }

    def __init__(self, args: Dict = {}, session: ASFSession = ASFSession()):
        super().__init__(args, session)

        bytes_mapping = {
            entry['Name']: {'bytes': entry['SizeInBytes'], 'format': entry['Format']}
            for entry in self.properties['bytes']
        }
        md5sum_mapping = {
            entry['Name']: entry['Checksum']['Value']
            for entry in self.properties['bytes']
        }

        self.properties['bytes'] = bytes_mapping
        self.properties['md5sum'] = md5sum_mapping

        self.properties['additionalUrls'] = self._get_additional_urls()
        self.properties['browse'] = [url for url in self._get_urls() if url.endswith('.png') or url.endswith('.jpg') or url.endswith('.jpeg')]
        self.properties['s3Urls'] = self._get_s3_uris()

        center = self.centroid()
        self.properties['centerLat'] = center.y
        self.properties['centerLon'] = center.x



import importlib.util
import math

from .ASFProduct import ASFProduct
from .baseline import calculate_perpendicular_baselines
from .exceptions import CoherenceEstimationError
import pytz

_COHERENCE_OPT_DEPS = ['zarr', 's3fs', 'rioxarray', 'xarray']
try:
    for spec in _COHERENCE_OPT_DEPS:
        if importlib.util.find_spec(spec) is None:
            raise ImportError

    import fsspec
    import xarray as xr

except ImportError:
    fsspec = None
    xr = None

try:
    from ciso8601 import parse_datetime
except ImportError:
    from dateutil.parser import parse as parse_datetime


class Pair:
    """
    A Pair is comprised of a reference scene and a secondary scene. These scenes typically intersect geographically,
    but that is not a requirement. When a pair is created, its perpendicular and temporal baselines are calculated
    and stored in the self.perpendicular_baseline and self.temporal_baseline member variables.

    Two pairs are equivalent if they have matching reference and secondary dates

    ref: ASFProduct for reference scene
    sec: ASFProduct for secondary scene
    """
    def __init__(self, ref: ASFProduct, sec: ASFProduct):
        self.ref = ref
        self.sec = sec
        self.id = (ref.properties['sceneName'], sec.properties['sceneName'])

        self.perpendicular_baseline = calculate_perpendicular_baselines(
            ref.properties['sceneName'], 
            [sec, ref])[0].properties['perpendicularBaseline']

        self.ref_time = parse_datetime(ref.properties["startTime"])
        if self.ref_time.tzinfo is None:
            self.ref_time = pytz.utc.localize(self.ref_time)
        self.sec_time = parse_datetime(sec.properties["startTime"])
        if self.sec_time.tzinfo is None:
            self.sec_time = pytz.utc.localize(self.sec_time)

        self.temporal_baseline = self.sec_time.date() - self.ref_time.date()

    def __repr__(self) -> str:
        return f"Pair({self.id[0]}, {self.id[1]})"

    def __eq__(self, other):
        if not isinstance(other, Pair):
            return NotImplemented
        return self.id == other.id

    def __hash__(self) -> int:
        return hash(self.id)

    def estimate_s1_mean_coherence(self) -> float:
        '''
        Estimates mean coherence for a Pair of Sentinel-1 scenes or bursts using the 11367x4367 overview of the 2019-2020 
        VV COH data from the Global Seasonal Sentinel-1 Interferometric Coherence and Backscatter Dataset:
        https://asf.alaska.edu/datasets/daac/global-seasonal-sentinel-1-interferometric-coherence-and-backscatter-dataset/

        To support effecient in-place subsetting and access, the VV COH data has been saved to a public Zarr Store in AWS S3:
        s3://asf-search-coh/global_coh_100ppd_11367x4367

        Returns:
        '''
        if xr is None or fsspec is None:
            raise ImportError(
                'The `estimate_s1_mean_coherence()` method requires the optional asf-search '
                f'dependencies {_COHERENCE_OPT_DEPS}, '
                'but they could not be found in the current python environment. '
                'Enable this method by including the appropriate pip or conda install. '
                'Ex: `python -m pip install asf-search[coherence]`'
            )

        month = parse_datetime(self.ref.properties["startTime"]).month
        if month in [12, 1, 2]:
            season = 'winter'
        elif month in [3, 4, 5]:
            season = 'spring'
        elif month in [6, 7, 8]:
            season = 'summer'
        elif month in [9, 10, 11]:
            season = 'fall'

        temporal = math.ceil(self.temporal_baseline.days / 6) * 6
        if temporal > 48:
            msg = (f"""Coherence dataset includes temporal baselines up to 48 days.
                   Temporal baseline: {self.temporal_baseline.days} days""")
            raise CoherenceEstimationError(msg)

        uri = f"s3://asf-search-coh/global_coh_100ppd_11367x4367_Zarrv2/Global_{season}_vv_COH{temporal}_100ppd.zarr"
        coords = self.ref.geometry['coordinates'][0]
        lons, lats = zip(*coords)
        minx, miny, maxx, maxy = min(lons), min(lats), max(lons), max(lats)

        ds = xr.open_zarr(
            fsspec.get_mapper(uri, s3={'anon': True}),
            consolidated=False
            )
        ds = ds.rio.write_crs("EPSG:4326", inplace=False)
        subset = ds.rio.clip_box(minx=minx, miny=miny, maxx=maxx, maxy=maxy)
        return subset.coherence.mean().compute().item()



from typing import Dict
from asf_search.exceptions import CMRError
from asf_search.constants.INTERNAL import CMR_HOST, CMR_COLLECTIONS_PATH

import requests


def get_campaigns(data) -> Dict:
    """Queries CMR Collections endpoint for
    collections associated with the given platform

    :param data: a dictionary with required keys:
    'include_facets', 'provider', 'platform[]' and optional key: 'instrument[]'

    :return: Dictionary containing CMR umm_json response
    """
    response = requests.post(f'https://{CMR_HOST}{CMR_COLLECTIONS_PATH}', data=data)
    if response.status_code != 200:
        raise CMRError(f'CMR_ERROR {response.status_code}: {response.text}')

    try:
        data = response.json()
    except Exception as e:
        raise CMRError(f'CMR_ERROR: Error parsing JSON from CMR: {e}')

    return data



import platform
from typing import List, Optional, Union
import requests
from requests.utils import get_netrc_auth
from requests.auth import HTTPBasicAuth
import http.cookiejar

from asf_search import ASF_LOGGER, __name__ as asf_name, __version__ as asf_version
from asf_search.exceptions import ASFAuthenticationError
import warnings


class ASFSession(requests.Session):
    def __init__(
        self,
        edl_host: Optional[str] = None,
        edl_client_id: Optional[str] = None,
        asf_auth_host: Optional[str] = None,
        cmr_host: Optional[str] = None,
        cmr_collections: Optional[str] = None,
        auth_domains: Optional[List[str]] = None,
        auth_cookie_names: Optional[List[str]] = None,
    ):
        """
        ASFSession is a subclass of `requests.Session`, and is meant to ease
        downloading ASF hosted data by simplifying logging in to Earthdata Login.

        To create an EDL account, see here: https://urs.earthdata.nasa.gov/users/new

        ASFSession provides three built-in methods for authorizing downloads:
        - EDL Username and Password: `auth_with_creds()`
        - EDL Token: `auth_with_token()`
        - Authenticated cookiejars: `auth_with_cookiejar()`

        Parameters
        ----------
        `edl_host`:
            the Earthdata login endpoint used by auth_with_creds().
            Defaults to `asf_search.constants.INTERNAL.EDL_HOST`
        `edl_client_id`:
            The Earthdata Login client ID for this package.
            Defaults to `asf_search.constants.INTERNAL.EDL_CLIENT_ID`
        `asf_auth_host`:
            the ASF auth endpoint.
            Defaults to `asf_search.constants.INTERNAL.ASF_AUTH_HOST`
        `cmr_host`:
            the base CMR endpoint to test EDL login tokens against.
            Defaults to `asf_search.constants.INTERNAL.CMR_HOST`
        `cmr_collections`:
            the CMR endpoint path login tokens will be tested against.
            Defaults to `asf_search.constants.INTERNAL.CMR_COLLECTIONS`
        `auth_domains`:
            the list of authorized endpoints that are allowed to pass auth credentials.
            Defaults to `asf_search.constants.INTERNAL.AUTH_DOMAINS`.
            Authorization headers WILL NOT be stripped from the session object
            when redirected through these domains.
        `auth_cookie_names`:
            the list of cookie names to use when verifying
            with `auth_with_creds()` & `auth_with_cookiejar()`

        More information on Earthdata Login can be found here:
        https://urs.earthdata.nasa.gov/documentation/faq
        """
        super().__init__()
        user_agent = '; '.join(
            [
                f'Python/{platform.python_version()}',
                f'{requests.__name__}/{requests.__version__}',
                f'{asf_name}/{asf_version}',
            ]
        )

        self.headers.update({'User-Agent': user_agent})  # For all hosts
        self.headers.update({'Client-Id': f'{asf_name}_v{asf_version}'})  # For CMR

        from asf_search.constants import INTERNAL

        self.edl_host = INTERNAL.EDL_HOST if edl_host is None else edl_host
        self.edl_client_id = INTERNAL.EDL_CLIENT_ID if edl_client_id is None else edl_client_id
        self.asf_auth_host = INTERNAL.ASF_AUTH_HOST if asf_auth_host is None else asf_auth_host
        self.cmr_collections = (
            INTERNAL.CMR_COLLECTIONS if cmr_collections is None else cmr_collections
        )
        self.auth_domains = INTERNAL.AUTH_DOMAINS if auth_domains is None else auth_domains
        self.auth_cookie_names = (
            INTERNAL.AUTH_COOKIES if auth_cookie_names is None else auth_cookie_names
        )

        self.cmr_host = INTERNAL.CMR_HOST if cmr_host is None else cmr_host

    def __eq__(self, other):
        return (
            self.auth == other.auth
            and self.headers == other.headers
            and self.cookies == other.cookies
        )

    def auth_with_creds(self, username: str, password: str):
        """
        Authenticates the session using EDL username/password credentials

        This method calls the /api/users/find_or_create_token endpoint to get an EDL token and sets the asf-urs cookie from an ASF auth host
        (default `cumulus.asf.alaska.edu`)
        Parameters
        ----------
        username:
            EDL username, see https://urs.earthdata.nasa.gov/
        password:
            EDL password, see https://urs.earthdata.nasa.gov/
        host:
            (optional): EDL host to log in to

        Returns
        ----------
        ASFSession
        """

        self.auth = HTTPBasicAuth(username, password)
        token = self._get_urs_access_token()
        self._set_asf_urs_cookie()
        
        ASF_LOGGER.info('EDL Bearer Token retreived and asf-urs cookie sucessfully set')
        self.auth = None
        self._update_edl_token(token=token)

        return self

    def auth_with_token(self, token: str):
        """
        Authenticates the session using an EDL Authorization: Bearer token

        Note: this does not set the asf-urs token, which is necessary for downloading certain products like SLC BURSTS 
        Parameters
        ----------
        token:
        EDL Auth Token for authenticated downloads, see https://urs.earthdata.nasa.gov/user_tokens

        Returns
        ----------
        ASFSession
        """
        oauth_authorization = (
            f'https://{self.edl_host}/oauth/tokens/user?client_id={self.edl_client_id}'
        )

        ASF_LOGGER.info(f'Authenticating EDL token against {oauth_authorization}')
        response = self.post(url=oauth_authorization, data={'token': token})

        if not 200 <= response.status_code <= 299:
            if not self._try_legacy_token_auth(token=token):
                raise ASFAuthenticationError('Invalid/Expired token passed')

        ASF_LOGGER.info('EDL token authentication successful')
        self._update_edl_token(token=token)

        return self

    def _get_urs_access_token(self) -> str:

        login_url = f'https://{self.edl_host}/api/users/find_or_create_token'
        ASF_LOGGER.info(f'Attempting to get account EDL Bearer token via "{login_url}"')
        response = self.post(login_url)
        
        try:
            response.raise_for_status()
        except Exception as e:
            raise ASFAuthenticationError(f'Failed to log in with provided credentials. Original Exception: {str(e)}')

        token = response.json().get('access_token')
        
        ASF_LOGGER.info('EDL Bearer Token retreived, using for future queries and downloads')
        return token

    def _set_asf_urs_cookie(self):
        """Verify login via edl host and redirect to ASF_AUTH_HOST to set asf-urs auth cookies (need this for SLC Burst extractor)"""
        login_url = f'https://{self.edl_host}/oauth/authorize?splash=false&client_id={self.edl_client_id}&response_type=code&redirect_uri=https://{self.asf_auth_host}/login'  # noqa F401

        ASF_LOGGER.info(f'Attempting to get "asf-urs" cookie via "{login_url}"')
        response = self.get(login_url)

        try:
            response.raise_for_status()

            if not self._check_auth_cookies(self.cookies.get_dict()):
                raise ASFAuthenticationError(f'Failed to find `asf-urs` cookie in response from {login_url}. Unable to perform certain searches')

            ASF_LOGGER.info(f'Authenticated {self.edl_host} against {self.asf_auth_host}, cookies set sucessfully')
        except Exception as e:
            raise ASFAuthenticationError(f'Failed to set asf-urs cookies. Original exception: {e}')

    def _try_legacy_token_auth(self, token: str) -> bool:
        """
        Checks `cmr_host` search endpoint directly with provided token
        using method used in previous versions of asf-search (<7.0.9).

        This may be removed in a future release
        """
        from asf_search.constants import INTERNAL

        if self.cmr_host != INTERNAL.CMR_HOST:
            self.headers.update({'Authorization': 'Bearer {0}'.format(token)})
            legacy_auth_url = f'https://{self.cmr_host}{self.cmr_collections}'
            response = self.get(legacy_auth_url)
            self.headers.pop('Authorization')
            return 200 <= response.status_code <= 299

        return False

    def _update_edl_token(self, token: str):
        self.headers.update({'Authorization': 'Bearer {0}'.format(token)})

    def auth_with_cookiejar(
        self,
        cookies: Union[http.cookiejar.CookieJar, requests.cookies.RequestsCookieJar],
    ):
        """
        Authenticates the session using a pre-existing cookiejar

        :param cookies: Any http.cookiejar compatible object

        :return ASFSession: returns self for convenience
        """
        if not self._check_auth_cookies(cookies):
            raise ASFAuthenticationError('Cookiejar does not contain login cookies')

        for cookie in cookies:
            if cookie.is_expired():
                raise ASFAuthenticationError('Cookiejar contains expired cookies')

        token = cookies.get_dict().get('urs-access-token')
        if token is None:
            ASF_LOGGER.warning(
                'Failed to find EDL Token in cookiejar. '
                'EDL Token expected in "urs-access-token" cookie, '
                'required for hidden/restricted dataset access.'
            )
        else:
            ASF_LOGGER.info('Authenticating EDL token found in "urs-access-token" cookie')
            try:
                self.auth_with_token(token)
            except ASFAuthenticationError:
                ASF_LOGGER.warning(
                    'Failed to authenticate with found EDL token found. '
                    'Access to hidden/restricted cmr data may be limited.'
                )

        self.cookies = cookies

        return self

    def _check_auth_cookies(
        self,
        cookies: Union[http.cookiejar.CookieJar, requests.cookies.RequestsCookieJar],
    ) -> bool:
        if isinstance(cookies, requests.cookies.RequestsCookieJar):
            cookies = dict(cookies)

        return any(cookie in self.auth_cookie_names for cookie in cookies)

    def rebuild_auth(self, prepared_request: requests.Request, response: requests.Response):
        """
        Overrides requests.Session.rebuild_auth()
        default behavior of stripping the Authorization header
        upon redirect. This allows token authentication to work with redirects to trusted domains
        """

        headers = prepared_request.headers
        url = prepared_request.url

        if 'Authorization' in headers:
            original_domain = '.'.join(self._get_domain(response.request.url).split('.')[-3:])
            redirect_domain = '.'.join(self._get_domain(url).split('.')[-3:])

            if original_domain != redirect_domain and (
                original_domain not in self.auth_domains or redirect_domain not in self.auth_domains
            ):
                del headers['Authorization']

        new_auth = get_netrc_auth(url) if self.trust_env else None
        if new_auth is not None:
            prepared_request.prepare_auth(new_auth)

    def _get_domain(self, url: str):
        return requests.utils.urlparse(url).hostname

    # multi-processing does an implicit copy of ASFSession objects,
    # this ensures ASFSession class variables are included
    def __getstate__(self):
        state = super().__getstate__()
        state = {
            **state,
            'edl_host': self.edl_host,
            'edl_client_id': self.edl_client_id,
            'asf_auth_host': self.asf_auth_host,
            'cmr_host': self.cmr_host,
            'cmr_collections': self.cmr_collections,
            'auth_domains': self.auth_domains,
            'auth_cookie_names': self.auth_cookie_names,
        }
        return state



class RepairEntry:
    def __init__(self, report_type: str, report: str) -> None:
        self.report_type = report_type
        self.report = report

    def __str__(self) -> str:
        return f'{self.report_type}: {self.report}'



field_map = {
    # API parameter                 CMR keyword                         CMR format strings
    'absoluteOrbit':                {'key': 'orbit_number',             'fmt': '{0}'},
    'asfFrame':                     {'key': 'attribute[]',              'fmt': 'int,FRAME_NUMBER,{0}'},
    'maxBaselinePerp':              {'key': 'attribute[]',              'fmt': 'float,INSAR_BASELINE,,{0}'},
    'minBaselinePerp':              {'key': 'attribute[]',              'fmt': 'float,INSAR_BASELINE,{0},'},
    'bbox':                         {'key': 'bounding_box',             'fmt': '{0}'},
    'beamMode':                     {'key': 'attribute[]',              'fmt': 'string,BEAM_MODE,{0}'},
    'beamSwath':                    {'key': 'attribute[]',              'fmt': 'string,BEAM_MODE_TYPE,{0}'},
    'campaign':                     {'key': 'attribute[]',              'fmt': 'string,MISSION_NAME,{0}'},
    'circle':                       {'key': 'circle',                   'fmt': '{0}'},
    'maxDoppler':                   {'key': 'attribute[]',              'fmt': 'float,DOPPLER,,{0}'},
    'minDoppler':                   {'key': 'attribute[]',              'fmt': 'float,DOPPLER,{0},'},
    'maxFaradayRotation':           {'key': 'attribute[]',              'fmt': 'float,FARADAY_ROTATION,,{0}'},  # noqa F401
    'minFaradayRotation':           {'key': 'attribute[]',              'fmt': 'float,FARADAY_ROTATION,{0},'},  # noqa F401
    'flightDirection':              {'key': 'attribute[]',              'fmt': 'string,ASCENDING_DESCENDING,{0}'},  # noqa F401
    'flightLine':                   {'key': 'attribute[]',              'fmt': 'string,FLIGHT_LINE,{0}'},
    'frame':                        {'key': 'attribute[]',              'fmt': 'int,CENTER_ESA_FRAME,{0}'},
    'granule_list':                 {'key': 'readable_granule_name[]',  'fmt': '{0}'},
    'groupID':                      {'key': 'attribute[]',              'fmt': 'string,GROUP_ID,{0}'},
    'insarStackId':                 {'key': 'attribute[]',              'fmt': 'int,INSAR_STACK_ID,{0}'},
    'linestring':                   {'key': 'line',                     'fmt': '{0}'},
    'lookDirection':                {'key': 'attribute[]',              'fmt': 'string,LOOK_DIRECTION,{0}'},
    'maxInsarStackSize':            {'key': 'attribute[]',              'fmt': 'int,INSAR_STACK_SIZE,,{0}'},
    'minInsarStackSize':            {'key': 'attribute[]',              'fmt': 'int,INSAR_STACK_SIZE,{0},'},
    'instrument':                   {'key': 'instrument[]',             'fmt': '{0}'},
    'offNadirAngle':                {'key': 'attribute[]',              'fmt': 'float,OFF_NADIR_ANGLE,{0}'},
    'platform':                     {'key': 'platform[]',               'fmt': '{0}'},
    'polarization':                 {'key': 'attribute[]',              'fmt': 'string,POLARIZATION,{0}'},
    'point':                        {'key': 'point',                    'fmt': '{0}'},
    'polygon':                      {'key': 'polygon',                  'fmt': '{0}'},
    'processingDate':               {'key': 'updated_since',            'fmt': '{0}'},
    'processingLevel':              {'key': 'attribute[]',              'fmt': 'string,PROCESSING_TYPE,{0}'},
    'product_list':                 {'key': 'granule_ur[]',             'fmt': '{0}'},
    'provider':                     {'key': 'provider',                 'fmt': '{0}'},
    'relativeOrbit':                {'key': 'attribute[]',              'fmt': 'int,PATH_NUMBER,{0}'},
    'temporal':                     {'key': 'temporal',                 'fmt': '{0}'},
    'collections':                  {'key': 'echo_collection_id[]',     'fmt': '{0}'},
    'shortName':                    {'key': 'shortName',                'fmt': '{0}'},
    'temporalBaselineDays':         {'key': 'attribute[]',              'fmt': 'int,TEMPORAL_BASELINE_DAYS,{0}'},  # noqa F401
    'ariaVersion':                  {'key': 'attribute[]',              'fmt': 'string,VERSION,{0}'},
    # SLC BURST fields
    'absoluteBurstID':              {'key': 'attribute[]',              'fmt': 'int,BURST_ID_ABSOLUTE,{0}'},
    'relativeBurstID':              {'key': 'attribute[]',              'fmt': 'int,BURST_ID_RELATIVE,{0}'},
    'fullBurstID':                  {'key': 'attribute[]',              'fmt': 'string,BURST_ID_FULL,{0}'},
    # OPERA-S1 field
    'operaBurstID':                 {'key': 'attribute[]',              'fmt': 'string,OPERA_BURST_ID,{0}'},
    # NISAR fields
    'mainBandPolarization':         {'key': 'attribute[]',             'fmt': 'string,FREQUENCY_A_POLARIZATION_CONCAT,{0}'},
    'sideBandPolarization':         {'key': 'attribute[]',             'fmt': 'string,FREQUENCY_B_POLARIZATION_CONCAT,{0}'},
    'frameCoverage':                {'key': 'attribute[]',             'fmt': 'string,FULL_FRAME,{0}'},
    'jointObservation':             {'key': 'attribute[]',             'fmt': 'string,JOINT_OBSERVATION,{0}'},
    'rangeBandwidth':               {'key': 'attribute[]',             'fmt': 'string,RANGE_BANDWIDTH_CONCAT,{0}'},
    'productionConfiguration':      {'key': 'attribute[]',             'fmt': 'string,PRODUCTION_PIPELINE,{0}'},
    'tileID':                       {'key': 'attribute[]',             'fmt': 'string,MGRS_TILE_ID,{0}'},
}



from copy import deepcopy
from unittest.mock import patch
from asf_search.ASFSearchOptions import ASFSearchOptions
from asf_search.Products import ARIAS1GUNWProduct
from asf_search.exceptions import ASFBaselineError, ASFSearchError
from asf_search.ASFSearchResults import ASFSearchResults
from asf_search import ASFSession, DATASET, BEAMMODE, POLARIZATION, PRODUCT_TYPE
from asf_search.search.baseline_search import stack_from_id, stack_from_product
from asf_search.baseline.stack import calculate_temporal_baselines
import pytest

from asf_search.search.search_generator import as_ASFProduct
from asf_enumeration import aria_s1_gunw


def run_test_get_preprocessed_stack_params(product):
    reference = as_ASFProduct(product, ASFSession())
    params = reference.get_stack_opts()

    original_properties = product['properties']

    assert params.processingLevel == [reference.get_default_baseline_product_type()]
    assert params.insarStackId == original_properties['insarStackId']
    assert len(dict(params)) == 2


def run_test_get_unprocessed_stack_params(product):
    reference = as_ASFProduct(product, ASFSession())
    params = reference.get_stack_opts()

    original_properties = product['properties']
    assert original_properties['polarization'] in params.polarization

    if reference.properties['processingLevel'] == 'BURST':
        assert [reference.properties['polarization']] == params.polarization
        assert [reference.properties['burst']['fullBurstID']] == params.fullBurstID
    elif reference.properties['sceneName'].startswith('S1-GUNW'):
        assert params.platform == ['SA', 'SB', 'SC', 'SD']
        assert DATASET.SENTINEL1 in params.dataset
        assert params.processingLevel == [PRODUCT_TYPE.SLC]
        assert params.beamMode == [BEAMMODE.IW]
        assert params.polarization == [POLARIZATION.VV, POLARIZATION.VV_VH]
        assert params.flightDirection.upper() == reference.properties['flightDirection'].upper()
        assert params.relativeOrbit == [reference.properties['pathNumber']]
    else:
        assert (
            ['VV', 'VV+VH'] == params.polarization
            if reference.properties['polarization'] in ['VV', 'VV+VH']
            else ['HH', 'HH+HV'] == params.polarization
        )
        assert len(dict(params)) == 7


def run_get_stack_opts_invalid_insarStackId(product):
    invalid_reference = as_ASFProduct(product, ASFSession())

    invalid_reference.properties['insarStackId'] = '0'

    with pytest.raises(ASFBaselineError):
        invalid_reference.get_stack_opts()


def run_test_calc_temporal_baselines(reference, stack):
    reference = as_ASFProduct(reference, ASFSession())
    stack = ASFSearchResults([as_ASFProduct(product, ASFSession()) for product in stack])
    stackLength = len(stack)

    calculate_temporal_baselines(reference, stack)

    assert len(stack) == stackLength
    for secondary in stack:
        assert 'temporalBaseline' in secondary.properties


def run_test_stack_from_product(reference, stack):
    reference = as_ASFProduct(reference, ASFSession())

    with patch('asf_search.baseline_search.search') as search_mock:
        search_mock.return_value = ASFSearchResults(
            [as_ASFProduct(product, ASFSession()) for product in stack]
        )

        stack = stack_from_product(reference)

        for idx, secondary in enumerate(stack):
            if idx > 0:
                assert (
                    secondary.properties['temporalBaseline']
                    >= stack[idx - 1].properties['temporalBaseline']
                )


def run_test_stack_from_id(stack_id: str, reference, stack, opts: ASFSearchOptions):
    temp = deepcopy(stack)

    with patch('asf_search.baseline_search.product_search') as mock_product_search:
        mock_product_search.return_value = ASFSearchResults(
            [as_ASFProduct(product, ASFSession()) for product in stack]
        )

        if not stack_id:
            with pytest.raises(ASFSearchError):
                stack_from_id(stack_id)
        else:
            with patch('asf_search.baseline_search.search') as search_mock:
                search_mock.return_value = ASFSearchResults(
                    [as_ASFProduct(product, ASFSession()) for product in temp]
                )

                returned_stack = stack_from_id(stack_id, opts=opts)
                stack_files = set(x['properties']['fileID'] for x in stack)
                filtered_stack = [
                    x for x in returned_stack if x.properties['fileID'] in stack_files
                ]
                for idx, secondary in enumerate(filtered_stack):
                    if idx > 0:
                        assert (
                            secondary.properties['temporalBaseline']
                            >= stack[idx - 1]['properties']['temporalBaseline']
                        )



from .S1Product import S1Product  # noqa: F401
from .ALOSProduct import ALOSProduct  # noqa: F401
from .RADARSATProduct import RADARSATProduct  # noqa: F401
from .AIRSARProduct import AIRSARProduct  # noqa: F401
from .ERSProduct import ERSProduct  # noqa: F401
from .JERSProduct import JERSProduct  # noqa: F401
from .UAVSARProduct import UAVSARProduct  # noqa: F401
from .SIRCProduct import SIRCProduct  # noqa: F401
from .SEASATProduct import SEASATProduct  # noqa: F401
from .SMAPProduct import SMAPProduct  # noqa: F401
from .S1BurstProduct import S1BurstProduct  # noqa: F401
from .OPERAS1Product import OPERAS1Product  # noqa: F401
from .ARIAS1GUNWProduct import ARIAS1GUNWProduct  # noqa: F401
from .NISARProduct import NISARProduct  # noqa: F401
from .ALOS2Product import ALOS2Product  # noqa: F401
from .TROPOProduct import TROPOProduct # noqa: F401



import inspect
from types import GeneratorType
from typing import Dict
from asf_search import ASF_LOGGER
from asf_search.export.metalink import MetalinkStreamArray
import xml.etree.ElementTree as ETree

extra_kml_fields = [
    (
        "configurationName",
        ["AdditionalAttributes", ("Name", "BEAM_MODE_DESC"), "Values", 0],
    ),
    (
        "faradayRotation",
        ["AdditionalAttributes", ("Name", "FARADAY_ROTATION"), "Values", 0],
    ),
    (
        "processingTypeDisplay",
        ["AdditionalAttributes", ("Name", "PROCESSING_TYPE_DISPLAY"), "Values", 0],
    ),
    ("sceneDate", ["AdditionalAttributes", ("Name", "ACQUISITION_DATE"), "Values", 0]),
    (
        "shape",
        [
            "SpatialExtent",
            "HorizontalSpatialDomain",
            "Geometry",
            "GPolygons",
            0,
            "Boundary",
            "Points",
        ],
    ),
    ("thumbnailUrl", ["AdditionalAttributes", ("Name", "THUMBNAIL_URL"), "Values", 0]),
    (
        "faradayRotation",
        ["AdditionalAttributes", ("Name", "FARADAY_ROTATION"), "Values", 0],
    ),
    (
        "offNadirAngle",
        ["AdditionalAttributes", ("Name", "OFF_NADIR_ANGLE"), "Values", 0],
    ),
]


def results_to_kml(results):
    ASF_LOGGER.info("Started translating results to kml format")

    if inspect.isgeneratorfunction(results) or isinstance(results, GeneratorType):
        return KMLStreamArray(results)

    return KMLStreamArray([results])


class KMLStreamArray(MetalinkStreamArray):
    def __init__(self, results):
        MetalinkStreamArray.__init__(self, results)
        self.header = """<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
 <kml xmlns="http://www.opengis.net/kml/2.2">
   <Document>
     <name>ASF Datapool Search Results</name>
     <description>Search Performed:</description>
     <Style id="yellowLineGreenPoly">
         <LineStyle>
         <color>30ff8800</color>
         <width>4</width>
         </LineStyle>
         <PolyStyle>
         <color>7f00ff00</color>
         </PolyStyle>
     </Style>\n     """
        self.footer = """</Document>\n</kml>"""

    def getOutputType(self) -> str:
        return "kml"

    def get_additional_fields(self, product):
        umm = product.umm
        additional_fields = {}
        for key, path in extra_kml_fields:
            additional_fields[key] = product.umm_get(umm, *path)
        return additional_fields

    def getItem(self, p):
        placemark = ETree.Element("Placemark")
        name = ETree.Element("name")
        name.text = p.get("sceneName")
        placemark.append(name)

        description = ETree.Element("description")
        description.text = """&lt;![CDATA["""
        placemark.append(description)

        h1 = ETree.Element("h1")
        h1.text = (
            f"{p.get('platform')} ({p.get('configurationName')}), acquired {p.get('sceneDate')}"
        )
        h2 = ETree.Element("h2")
        h2.text = p.get("url", "")
        description.append(h1)
        description.append(h2)

        div = ETree.Element(
            "div", attrib={"style": "position:absolute;left:20px;top:200px"}
        )
        description.append(div)

        h3 = ETree.Element("h3")
        h3.text = "Metadata"
        div.append(h3)

        ul = ETree.Element("ul")
        div.append(ul)

        for text, value in self.metadata_fields(p).items():
            li = ETree.Element("li")
            li.text = text + str(value)
            ul.append(li)

        if p.get('platform') == 'NISAR':
            h3.text = "Files"
            div.append(h3)
            ul_files = ETree.Element("ul")
            div.append(ul_files)
            for url in p.get('additionalUrls'):
                li = ETree.Element("li")
                li.text = url
                ul_files.append(li)

        d = ETree.Element(
            "div", attrib={"style": "position:absolute;left:300px;top:250px"}
        )
        description.append(d)

        a = ETree.Element("a")
        if p.get("browse") is not None and len(p.get("browse")):
            a.set("href", p.get("browse")[0])
        else:
            a.set("href", "")

        d.append(a)

        img = ETree.Element("img")
        if p.get("thumbnailUrl") is not None:
            img.set("src", p.get("thumbnailUrl"))
        else:
            img.set("src", "None")
        a.append(img)

        styleUrl = ETree.Element("styleUrl")
        styleUrl.text = "#yellowLineGreenPoly"
        placemark.append(styleUrl)

        polygon = ETree.Element("Polygon")
        placemark.append(polygon)

        extrude = ETree.Element("extrude")
        extrude.text = "1"
        polygon.append(extrude)

        altitudeMode = ETree.Element("altitudeMode")
        altitudeMode.text = "relativeToGround"
        polygon.append(altitudeMode)

        outerBondaryIs = ETree.Element("outerBoundaryIs")
        polygon.append(outerBondaryIs)

        linearRing = ETree.Element("LinearRing")
        outerBondaryIs.append(linearRing)

        coordinates = ETree.Element("coordinates")

        if p.get("shape") is not None:
            coordinates.text = (
                "\n"
                + (14 * " ")
                + ("\n" + (14 * " ")).join(
                    [f"{c['Longitude']},{c['Latitude']},2000" for c in p.get("shape")]
                )
                + "\n"
                + (14 * " ")
            )
        linearRing.append(coordinates)

        self.indent(placemark, 3)

        # for CDATA section, manually replace &amp; escape character with &
        return ETree.tostring(placemark, encoding="unicode").replace("&amp;", "&")

    # Helper method for getting additional fields in <ul> tag
    def metadata_fields(self, item: Dict):
        required = {
            'Processing type: ': item.get('processingTypeDisplay'),
            'Frame: ': item.get('frameNumber'),
            'Path: ': item.get('pathNumber'),
            'Orbit: ': item.get('orbit'),
            'Start time: ': item.get('startTime'),
            'End time: ': item.get('stopTime'),
        }

        optional = {}
        for text, key in [
            ("Faraday Rotation: ", "faradayRotation"),
            ("Ascending/Descending: ", "flightDirection"),
            ("Off Nadir Angle: ", "offNadirAngle"),
            ("Pointing Angle: ", "pointingAngle"),
            ("Temporal Baseline: ", "temporalBaseline"),
            ("Perpendicular Baseline: ", "perpendicularBaseline"),
        ]:
            if item.get(key) is not None:
                if isinstance(item[key], float) and key == "offNadirAngle":
                    optional[text] = f"{item[key]:g}"  # trim trailing zeros
                else:
                    optional[text] = item[key]
            elif key not in ["temporalBaseline", "perpendicularBaseline"]:
                optional[text] = "None"

        output = {**required, **optional}
        if item["processingLevel"] == "BURST":
            burst = {
                "Absolute Burst ID: ": item["burst"]["absoluteBurstID"],
                "Relative Burst ID: ": item["burst"]["relativeBurstID"],
                "Full Burst ID: ": item["burst"]["fullBurstID"],
                "Burst Index: ": item["burst"]["burstIndex"],
                "Azimuth Time: ": item["burst"]["azimuthTime"],
                "Azimuth Anx Time: ": item["burst"]["azimuthAnxTime"],
                "Samples per Burst: ": item["burst"]["samplesPerBurst"],
                "Subswath: ": item["burst"]["subswath"],
            }

            output = {**output, **burst}

        return output



from datetime import datetime, timedelta, timezone
import pickle
from pathlib import Path

from asf_search import Pair
import numpy as np

import pytest

@pytest.fixture
def reference_full_scene():
    path = Path(__file__).parent / "data/reference_full_scene_0.pkl"

    with open(path, "rb") as f:
        # trusted local test fixture
        return pickle.load(f) # nosec B301
    
@pytest.fixture
def secondary_full_scene():
    path = Path(__file__).parent / "data/secondary_full_scene_0.pkl"

    with open(path, "rb") as f:
        # trusted local test fixture
        return pickle.load(f) # nosec B301
    
@pytest.fixture
def reference_burst():
    path = Path(__file__).parent / "data/reference_burst_0.pkl"

    with open(path, "rb") as f:
        # trusted local test fixture
        return pickle.load(f) # nosec B301
    
@pytest.fixture
def secondary_burst():
    path = Path(__file__).parent / "data/secondary_burst_0.pkl"

    with open(path, "rb") as f:
        # trusted local test fixture
        return pickle.load(f) # nosec B301


def test_make_full_scene_pair(reference_full_scene, secondary_full_scene):
    full_scene_pair = Pair(reference_full_scene, secondary_full_scene)
    assert full_scene_pair.ref.properties['sceneName'] == "S1A_IW_SLC__1SDV_20220215T225119_20220215T225146_041930_04FE2E_9252"
    assert full_scene_pair.sec.properties['sceneName'] == "S1A_IW_SLC__1SDV_20220227T225119_20220227T225146_042105_050431_987E"
    assert full_scene_pair.ref_time == datetime(2022, 2, 15, 22, 51, 19, tzinfo=timezone.utc)
    assert full_scene_pair.sec_time == datetime(2022, 2, 27, 22, 51, 19, tzinfo=timezone.utc)
    assert full_scene_pair.perpendicular_baseline == -15
    assert full_scene_pair.temporal_baseline == timedelta(days=12)
    assert np.floor(full_scene_pair.estimate_s1_mean_coherence()) == 18.0

def test_make_burst_pair(reference_burst, secondary_burst):
    burst_pair = Pair(reference_burst, secondary_burst)
    assert burst_pair.ref.properties['sceneName'] == "S1_181296_IW1_20220219T125501_VV_10AF-BURST"
    assert burst_pair.sec.properties['sceneName'] == "S1_181296_IW1_20220303T125501_VV_F03A-BURST"
    assert burst_pair.ref_time == datetime(2022, 2, 19, 12, 55, 3, tzinfo=timezone.utc)
    assert burst_pair.sec_time == datetime(2022, 3, 3, 12, 55, 2, tzinfo=timezone.utc)
    assert burst_pair.perpendicular_baseline == -75
    assert burst_pair.temporal_baseline == timedelta(days=12)
    assert np.floor(burst_pair.estimate_s1_mean_coherence()) == 52.0    



from datetime import datetime, timezone
from typing import Any, Dict, List, Optional
from asf_search.ASFSearchOptions import ASFSearchOptions
from asf_search.CMR.datasets import get_concept_id_alias
from asf_search.constants import CMR_PAGE_SIZE
import re
from shapely import wkt
from shapely.geometry import Polygon
from shapely.geometry.base import BaseGeometry
from .field_map import field_map
from .datasets import collections_per_platform, NISAR_PRODUCT_TYPES
import logging

try:
    from ciso8601 import parse_datetime
except ImportError:
    from dateutil.parser import parse as parse_datetime

nisar_collections_set = set(collections_per_platform['NISAR'])


def translate_opts(opts: ASFSearchOptions) -> List:
    # Need to add params which ASFSearchOptions cant support (like temporal),
    # so use a dict to avoid the validate_params logic:
    dict_opts = dict(opts)

    should_use_track = not set(dict_opts.get('collections', [])).isdisjoint(nisar_collections_set)

    if (
        dict_opts.get('processingLevel') is not None
    ):  # Certain products are now using PRODUCT_TYPE instead of PROCESSING_LEVEL
        processingType = dict_opts.get('processingLevel', [])[0]
        if processingType in NISAR_PRODUCT_TYPES:
            should_use_track = True
            # Use new PRODUCT_TYPE keyword later, remove processingLevel so we don't try the value with PROCESSING_LEVEL
            dict_opts['productType'] = dict_opts.pop('processingLevel')[0]

    # Escape commas for each key in the list.
    # intersectsWith, temporal, and other keys you don't want to escape, so keep whitelist instead
    for escape_commas in ['campaign']:
        if escape_commas in dict_opts:
            dict_opts[escape_commas] = dict_opts[escape_commas].replace(',', '\\,')

    dict_opts = fix_cmr_shapes(dict_opts)

    # Additional Attribute FULL_FRAME stored as a TRUE/FALSE string
    if 'frameCoverage' in dict_opts:
        dict_opts['frameCoverage'] = {
            'F': 'TRUE',
            'P': 'FALSE',
        }[dict_opts['frameCoverage'][0].upper()]

    if 'jointObservation' in dict_opts:
        dict_opts['jointObservation'] = str(dict_opts['jointObservation']).upper()

    # Special case to unravel WKT field a little for compatibility
    if 'intersectsWith' in dict_opts:
        shape = wkt.loads(dict_opts.pop('intersectsWith', None))

        # If a wide rectangle is provided, make sure to use the bounding box
        # instead of the wkt for better responses from CMR
        # This will provide better results with AOI's near poles
        if should_use_bbox(shape):
            bounds = shape.boundary.bounds
            if bounds[0] > 180 or bounds[2] > 180:
                bounds = [
                    (x + 180) % 360 - 180 if idx % 2 == 0 and abs(x) > 180 else x
                    for idx, x in enumerate(bounds)
                ]

            bottom_left = [str(coord) for coord in bounds[:2]]
            top_right = [str(coord) for coord in bounds[2:]]

            bbox = ','.join([*bottom_left, *top_right])
            dict_opts['bbox'] = bbox
        else:
            (shapeType, shape) = wkt_to_cmr_shape(shape).split(':')
            dict_opts[shapeType] = shape

    # If you need to use the temporal key:
    if any(key in dict_opts for key in ['start', 'end', 'season']):
        dict_opts = fix_date(dict_opts)

    dict_opts = fix_range_params(dict_opts)

    # convert the above parameters to a list of key/value tuples
    cmr_opts = []

    # user provided umm fields
    custom_cmr_keywords = dict_opts.pop('cmr_keywords', [])

    granule_list_with_wildcard = False
    for key, val in dict_opts.items():
        # If it's "session" or something else CMR doesn't accept, don't send it:
        if key not in field_map:
            if key == 'productType':
                custom_cmr_keywords.append(('attribute[]', f'string,PRODUCT_TYPE,{val}'))

            continue
        if isinstance(val, list):
            for x in val:
                if key in ['granule_list', 'product_list']:
                    for y in x.split(','):
                        cmr_opts.append((key, y))
                        if key == 'granule_list' and ('*' in y or '?' in y):
                            granule_list_with_wildcard = True
                else:
                    if isinstance(x, tuple):
                        cmr_opts.append((key, ','.join([str(t) for t in x])))
                    else:
                        cmr_opts.append((key, x))
        else:
            cmr_opts.append((key, val))
    # translate the above tuples to CMR key/values
    for i, opt in enumerate(cmr_opts):
        cmr_opts[i] = field_map[opt[0]]['key'], field_map[opt[0]]['fmt'].format(opt[1])

    if should_use_asf_frame(cmr_opts):
        cmr_opts = use_asf_frame(cmr_opts)

    if should_use_track:
        cmr_opts = use_track_number(cmr_opts)

    cmr_opts.extend(custom_cmr_keywords)

    additional_keys = [
        ('page_size', CMR_PAGE_SIZE),
        ('options[temporal][and]', 'true'),
        ('sort_key[]', '-end_date'),
        ('sort_key[]', 'granule_ur'),
        ('options[platform][ignore_case]', 'true'),
        ('provider', opts.provider),
    ]

    if granule_list_with_wildcard:
        additional_keys.append(('options[readable_granule_name][pattern]', 'true'))

    cmr_opts.extend(additional_keys)

    return cmr_opts


def fix_cmr_shapes(fixed_params: Dict[str, Any]) -> Dict[str, Any]:
    """Fixes raw CMR lon lat coord shapes"""
    for param in ['point', 'linestring', 'circle', 'bbox']:
        if param in fixed_params:
            fixed_params[param] = ','.join(map(str, fixed_params[param]))

    return fixed_params


def should_use_asf_frame(cmr_opts):
    asf_frame_platforms = [
        'SENTINEL-1A',
        'SENTINEL-1B',
        'SENTINEL-1C',
        'SENTINEL-1D',
        'ALOS',
        'ALOS-2',
        'NISAR',
        'SEASAT 1',
    ]

    asf_frame_collections = get_concept_id_alias(asf_frame_platforms, collections_per_platform)

    return any(
        [
            p[0] == 'platform[]'
            and p[1].upper() in asf_frame_platforms
            or p[0] == 'echo_collection_id[]'
            and p[1] in asf_frame_collections
            for p in cmr_opts
        ]
    )


def use_asf_frame(cmr_opts):
    """
    Sentinel/ALOS: always use asf frame instead of esa frame

    Platform-specific hack
    We do them at the subquery level in case the main query crosses
    platforms that don't suffer these issue.
    """

    for n, p in enumerate(cmr_opts):
        if not isinstance(p[1], str):
            continue

        m = re.search(r'CENTER_ESA_FRAME', p[1])
        if m is None:
            continue

        logging.debug('Sentinel/ALOS subquery, using ASF frame instead of ESA frame')

        cmr_opts[n] = (p[0], p[1].replace(',CENTER_ESA_FRAME,', ',FRAME_NUMBER,'))

    return cmr_opts


def use_track_number(cmr_opts):
    """
    NISAR: always use track number instead of path number
    """

    for n, p in enumerate(cmr_opts):
        if not isinstance(p[1], str):
            continue

        m = re.search(r'PATH_NUMBER', p[1])
        if m is None:
            continue

        logging.debug(
            'NISAR subquery with relativeOrbit, using TRACK_NUMBER instead of PATH_NUMBER'
        )

        cmr_opts[n] = (p[0], p[1].replace(',PATH_NUMBER,', ',TRACK_NUMBER,'))

    return cmr_opts


# some products don't have integer values in BYTES fields, round to nearest int
def try_round_float(value: str) -> Optional[int]:
    if value is None:
        return None

    value = float(value)
    return round(value)


def try_parse_int(value: str) -> Optional[int]:
    if value is None:
        return None

    return int(value)


def try_parse_float(value: str) -> Optional[float]:
    if value is None:
        return None

    return float(value)


def try_parse_bool(val: str) -> Optional[bool]:
    """Boolean values are stored as strings in umm json"""
    if val is None:
        return None

    return val.lower() == 'true'


def try_parse_frame_coverage(val: str) -> Optional[str]:
    """Frame Coverage is stored as a string boolean in FULL_FRAME, convert it to Partial/Full"""
    if val is not None:
        if val.lower() == 'true':
            val = 'Full'
        else:
            val = 'Partial'

    return val


def try_parse_date(value: str) -> Optional[str]:
    if value is None:
        return None

    try:
        date = parse_datetime(value)
    except ValueError:
        return None

    if date is None:
        return value

    if date.tzinfo is None:
        date = date.replace(tzinfo=timezone.utc)
        # Turn all inputs into a consistant format:

    return date.strftime('%Y-%m-%dT%H:%M:%SZ')


def fix_date(fixed_params: Dict[str, Any]):
    if 'start' in fixed_params or 'end' in fixed_params or 'season' in fixed_params:
        fixed_params['start'] = (
            fixed_params['start'] if 'start' in fixed_params else '1978-01-01T00:00:00Z'
        )
        fixed_params['end'] = (
            fixed_params['end'] if 'end' in fixed_params else datetime.now(timezone.utc).isoformat()
        )
        fixed_params['season'] = (
            ','.join(str(x) for x in fixed_params['season']) if 'season' in fixed_params else ''
        )

        fixed_params['temporal'] = (
            f'{fixed_params["start"]},{fixed_params["end"]},{fixed_params["season"]}'
        )

        # And a little cleanup
        fixed_params.pop('start', None)
        fixed_params.pop('end', None)
        fixed_params.pop('season', None)

    return fixed_params


def fix_range_params(fixed_params: Dict[str, Any]) -> Dict[str, Any]:
    """Converts ranges to comma separated strings"""
    for param in [
        'offNadirAngle',
        'relativeOrbit',
        'absoluteOrbit',
        'frame',
        'asfFrame',
    ]:
        if param in fixed_params.keys() and isinstance(fixed_params[param], list):
            fixed_params[param] = ','.join([str(val) for val in fixed_params[param]])

    return fixed_params


def should_use_bbox(shape: BaseGeometry):
    """
    If the passed shape is a polygon, and if that polygon
    is equivalent to it's bounding box (if it's a rectangle),
    we should use the bounding box to search instead
    """
    if isinstance(shape, Polygon):
        coords = [
            [shape.bounds[0], shape.bounds[1]],
            [shape.bounds[2], shape.bounds[1]],
            [shape.bounds[2], shape.bounds[3]],
            [shape.bounds[0], shape.bounds[3]],
        ]
        return shape.equals(Polygon(shell=coords))

    return False


def wkt_to_cmr_shape(shape: BaseGeometry):
    # take note of the WKT type
    if shape.geom_type not in ['Point', 'LineString', 'Polygon']:
        raise ValueError('Unsupported WKT: {0}.'.format(shape.wkt))

    if shape.geom_type == 'Polygon':
        coords = shape.exterior.coords
    else:  # type == Point | Linestring
        coords = shape.coords
    # Turn [[x,y],[x,y]] into [x,y,x,y]:
    lon_lat_sequence = []
    for lon_lat in coords:
        lon_lat_sequence.extend(lon_lat)
    # Turn any "6e8" to a literal number. (As a sting):
    coords = ['{:.16f}'.format(float(cord)) for cord in lon_lat_sequence]
    return '{0}:{1}'.format(shape.geom_type.lower(), ','.join(coords))
